#!/usr/bin/env python

""" metano Flux Balance Analysis

This script performs flux balance analysis (FBA) on a metabolic network
(given as a set of reaction equations), an objective function, and a number
of inequality constraints.

It parses a reaction file and a scenario file, formulates the
corresponding linear, quadratic, or other nonlinear optimization problem,
and calls an appropriate solver.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from builtins import range
from builtins import object
import optparse
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.metabolicmodel import MetabolicModel
from metano.metabolicflux import MetabolicFlux
from metano.linearproblem import LinearProblem, _GLPK, _OOGLPK
from metano.nonlinearproblem import NonLinearProblem
from metano.objectivefun import biomass_per_flux, neg_grad_biomass_per_flux
from metano.startpointiterator import StartPointIterator
from metano.defines import FbaParam, printHistogram, COPYRIGHT_VERSION_STRING
from numpy import array, insert, vstack, dot
import os


EPSILON = 1E-8


class OptionParser(optparse.OptionParser):
    """ Extension of OptionParser with function for checking required arguments
    """

    def check_required(self, opt):
        option = self.get_option(opt)

        # Assumes the option's 'default' is set to None!
        if getattr(self.values, option.dest) is None:
            self.error("%s option not supplied" % option)


class FbAnalyzer(object):
    """ Class for flux-balance analysis
    """

    def __init__(self, solver="default", verbose=True):
        """ initialize FBA class

        Keyword arguments:

        solver    -- name of the solver to be used for the LP/QP/NLP
        """
        self.solver = solver
        self.verb = verbose

    @staticmethod
    def splitFluxes(matrix, reaction_names, lb, ub):
        """ split the fluxes into non-negative components

        This is done by duplicating and negating matrix columns for reversible
        reactions and just negating the matrix columns corresponding to flipped
        irreversible reactions.

        Keyword arguments:

        matrix         -- the stoichiometric matrix
        reaction_names -- list of reactions (indexed like matrix columns)
        lb             -- list of lower bounds (indexed like matrix columns)
        ub             -- list of upper bounds (indexed like matrix columns)

        Returns:
        matrixSplit, reactionsSplit, lbSplit, ubSplit

        matrixSplit    -- matrix with extra columns for negative flux components
        reactionsSplit -- dictionary { name : pair of indexes } - indexes can be
                          None; if both are not None, the first is the pos. and
                          the second is the neg. component (v = v[0] - v[1])
        lbSplit        -- list of lower bounds (indexed like columns of
                                                matrixSplit)
        ubSplit        -- list of upper bounds ( - '' - )
        """

        matrixSplit = matrix.copy()     # Perform deep copy
        reactionsSplit = {}
        lbSplit = []
        ubSplit = []
        newIndex = 0                    # Index in matrixSplit

        for i in range(len(reaction_names)):
            rea = reaction_names[i]
            if lb[i] >= 0.:
                # Reaction is irreversible (pos. flux only or restricted to 0)
                reactionsSplit[rea] = (newIndex, None)
                lbSplit.append(lb[i])
                ubSplit.append(ub[i])
            elif ub[i] <= 0.:
                # Reaction is irreversible & flipped => negate
                reactionsSplit[rea] = (None, newIndex)
                matrixSplit[:, newIndex] *= -1
                lbSplit.append(-ub[i])
                ubSplit.append(-lb[i])
            else:
                # Reaction is reversible => split (copy, then negate the copy)
                reactionsSplit[rea] = (newIndex, newIndex+1)
                matrixSplit = insert(matrixSplit, newIndex+1,
                                     -matrixSplit[:, newIndex], axis=1)
                lbSplit.append(0.)
                ubSplit.append(ub[i])
                lbSplit.append(0.)
                ubSplit.append(-lb[i])
                newIndex += 1

            newIndex += 1

        return matrixSplit, reactionsSplit, lbSplit, ubSplit

    @staticmethod
    def splitFluxVector(vec, reaction_names, reactionsSplit):
        """ split the given vector as described by reactionsSplit

        Keyword arguments:

        vec            -- the flux vector to be split
        reaction_names -- list of reactions corresponding to entries of vec
        reactionsSplit -- dictionary { name : tuple of indexes } - indexes can
                          be None; if both are not None, the first is the pos.
                          and the second is the neg. component (v = v[0] - v[1])
                          (as computed by FbAnalyzer.splitFluxes)

        Returns:  split vector (containing only non-negative values)
        """
        vecSplit = []
        for i in range(len(reaction_names)):
            rea = reaction_names[i]
            indexPos, indexNeg = reactionsSplit[rea]
            if len(reactionsSplit[rea]) != 2:
                raise TypeError("reactionsSplit tuples must exactly 2 entries")
            if vec[i] == 0.:
                if indexPos is not None:
                    vecSplit.append(0.)
                if indexNeg is not None:
                    vecSplit.append(0.)
            elif vec[i] > 0.:
                if indexPos is None:
                    print(rea, indexPos, indexNeg)
                    raise ValueError("Value %u (%g) out of range in split "
                                     "vector (must not be positive)" % (i, vec[i]))
                vecSplit.append(vec[i])
                if indexNeg is not None:
                    vecSplit.append(0.)
            else:
                if indexPos is not None:
                    vecSplit.append(0.)
                if indexNeg is None:
                    raise ValueError("Value %u (%g) out of range in split "
                                     "vector (must not be negative)" % (i, vec[i]))
                vecSplit.append(-vec[i])

        return vecSplit

    @staticmethod
    def rejoinFluxes(solutionSplit, reactionsSplit, reactions):
        """ rejoin the fluxes in the given solution by applying reactionsSplit

        This assumes that solutionSplit has been obtained by performing FBA on
        the split matrix corresponding to reactionsSplit. This function combines
        the split fluxes by subtracting the negative component from the positive
        one.

        Keyword arguments:

        solutionSplit  -- vector of non-negative fluxes
        reactionsSplit -- dictionary { name : tuple of indexes } - indexes can
                          be None; if both are not None, the first is the pos.
                          and the second is the neg. component (v = v[0] - v[1])
                          (as computed by FbAnalyzer.splitFluxes)
        reactions      -- dictionary of all reactions { name : original matrix
                                                                        column }

        Returns:  joined flux vector (indexed like original matrix columns)
        """
        if len(solutionSplit) == 0:
            return solutionSplit  # Nothing to do
        solution = [0.]*len(reactions)

        for rea in reactionsSplit:
            rea_index_orig = reactions[rea]
            indexPos, indexNeg = reactionsSplit[rea]

            if indexPos is not None:
                solution[rea_index_orig] += solutionSplit[indexPos]

            if indexNeg is not None:
                solution[rea_index_orig] -= solutionSplit[indexNeg]

        return solution

    @staticmethod
    def makeConstraintMatrices(matrix, eqs, ineqs):
        """ transform the given matrix and equality and inequality constraints
            to left-hand-side matrices and right-hand-side vectors of equality
            and inequality constraints

        Keyword arguments:

        matrix -- the stoichiometric matrix of the metabolic network
        eqs    -- list of (coefficient vector, right-hand side) pairs for the
                  equality constraints
        ineqs  -- list of - '' - for the inequality constraints

        """
        if eqs:
            Aeq = vstack((matrix, array([row[0] for row in eqs])))
            beq = array([0.]*len(matrix) + [row[1] for row in eqs])
        else:
            Aeq = matrix
            beq = array([0.]*len(matrix))
        if ineqs:
            Aineq = array([row[0] for row in ineqs])
            bineq = array([row[1] for row in ineqs])
        else:
            Aineq, bineq = None, None

        return Aeq, beq, Aineq, bineq

    def run(self, reactions, matrix, lb, ub, fbaParams):
        """ analyze objective function, and solve the LP/QP/NLP

        Keyword arguments:

        reactions  -- dictionary of all reactions { name : matrix column } or
                      { name : tuple of indices } for split fluxes
        matrix    -- the stoichiometric matrix of the metabolic network
        lb        -- list of lower bounds (indexed like matrix columns)
        ub        -- list of upper bounds (indexed like matrix columns)
        fbaParams -- optimization parameters (incl. objective function)

        Returns:
        obj_value, solution

        obj_value -- optimal value of objective function
        solution  -- a solution where the objective function assumes obj_value
        """
        maxmin, objStr, numIter = (fbaParams.maxmin, fbaParams.objStr,
                                   fbaParams.numIter)
        # Multiply by -1 for maximization
        maxmin_factor = -1. if maxmin == True else 1.
        nCols = len(lb)

        # Get additional linear equality and inequality constraints
        try:
            Aeq, beq, Aineq, bineq = self.makeConstraintMatrices(matrix,
                                                                 *ParamParser.linConstraintsToVectors(fbaParams.linConstraints,
                                                                                                      reactions, nCols))
        except ValueError:
            # If any linear constraint is contradictory, return empty solution
            return 0., []

        if objStr.lower().startswith("per_flux"):
            # special nonlinear objective function: single flux per flux unit
            # e.g. "per_flux(Biomass)"
            try:
                str_perflux, str_flux = objStr.split('(', 2)
            except ValueError:
                print ("Error in scenario file in objective function "
                       "definition.\nPER_FLUX syntax: "
                       "PER_FLUX(<reaction_name>)")
                exit()

            if str_perflux.strip().lower() != "per_flux":
                print ("Error: Objective function '%s' is not a reaction name "
                       "or a PER_FLUX definition.\n"
                       "Currently, objective function must be a linear "
                       "combination of reaction fluxes or PER_FLUX(<reaction>)."
                       % objStr)
                exit()

            rea_name = str_flux.split(')', 1)[0].strip()
            try:
                biomass_index = reactions[rea_name][0]
            except TypeError:
                biomass_index = reactions[rea_name]

            def obj_func(x): return -biomass_per_flux(x, biomass_index)

            if numIter < 0:
                numIter = FbaParam.DEFAULT_NUMITER

            try:
                ps = NonLinearProblem(Aeq, beq, Aineq, bineq, lb, ub,
                                      self.solver)
                ps.setObjective(obj_func)
                ps.setObjGrad(lambda x:
                              neg_grad_biomass_per_flux(x, biomass_index))
                if fbaParams.nlc != []:
                    ps.setNonlinearConstraints(fbaParams.nlc)
                    ps.setNlcGrad(fbaParams.nlc_grad)
                spIter = StartPointIterator(nCols, numIter)
                spIter.setRange(-1., 1.)
                ps.setStartPointIterator(spIter)

            except ValueError as strerror:
                print(strerror)
                exit()

        else:
            # Build linear objective function (given as coefficient vector)
            try:
                objective = ParamParser.convertObjFuncToLinVec(objStr,
                                                               reactions, nCols, maxmin)
            except Exception as strerror:
                print ("Error while trying to build coefficient vector of "
                       "linear objective function:")
                print(strerror)
                exit()

            if fbaParams.nlc == []:
                # Linear function and linear constraints only

                # If flux variables are not split, use GLPK through OpenOpt
                solver = (_OOGLPK if self.solver.lower() == _GLPK and
                          nCols == len(reactions) else self.solver)
                try:
                    ps = LinearProblem(Aeq, beq, Aineq, bineq, lb, ub, solver)
                except ValueError as strerror:
                    print(strerror)
                    exit()
                #print "lb:", len(lb)
                ps.setObjective(objective)
            else:
                # Linear function but nonlinear constraints
                try:
                    ps = NonLinearProblem(Aeq, beq, Aineq, bineq, lb, ub,
                                          self.solver)
                except ValueError as strerror:
                    print(strerror)
                    exit()

                ps.setObjective(lambda x: dot(x, objective))
                ps.setObjGrad(lambda _: array(objective))
                ps.setNonlinearConstraints(fbaParams.nlc)
                ps.setNlcGrad(fbaParams.nlc_grad)
                spIter = StartPointIterator(nCols, numIter)
                spIter.setRange(-1., 1.)
                ps.setStartPointIterator(spIter)

        # Launch optimization problem solver
        try:
            obj_value, solution = ps.minimize()
            print(obj_value)
        except ValueError as strerror:
            if self.solver == "default":
                print("Default solver reported an error:")
            else:
                print("Solver %s reported an error:" % self.solver)
            print(strerror)
            print("Consider trying a different solver.")
            exit()
        except KeyError as strerror:
            print(strerror)
            exit()

        obj_value *= maxmin_factor
        if hasattr(ps, "getHistogram"):
            printHistogram(*ps.getHistogram(32)+(maxmin_factor < 0.,))

        return obj_value, solution

    def runWithTotalFluxMinimization(self, reactions, matrix, lb, ub, fbaParams,
                                     start_value_total=None, tolerance_total=EPSILON, tolerance_obj=EPSILON):
        """ perform FBA with LP iteratively to find a flux distribution with
            optimal value of objective function value and minimum total flux

        This only works with non-negative flux variables, i.e. split fluxes!
        The objective function and all constraints must be linear.

        This function limits the total flux to a parameter and then iteratively
        fits this parameter until either the change in total flux is less than
        tolerance_total or the change in the objective function is less than
        tolerance_obj

        Keyword arguments:

        reactions  -- dictionary { name : tuple of indices } for split fluxes
        matrix    -- the stoichiometric matrix of the metabolic network
        lb        -- list of lower bounds (indexed like matrix columns)
        ub        -- list of upper bounds (indexed like matrix columns)
        fbaParams -- optimization parameters (incl. objective function)
        start_value_total -- initial flux limit (if None: set to len(reactions))
        tolerance_total   -- stopping criterion: change in flux limit below this
        tolerance_obj     -- stopping criterion: change in objective function
                                                 value below this

        Returns:
        limit, obj_value, solution, nSteps

        limit     -- limit of total flux
        obj_value -- optimal value of objective function
        solution  -- a solution where the objective function assumes obj_value,
                     obtained with limited total flux
        nSteps    -- number of steps until optimal limit was found
        """
        maxmin, objStr = (fbaParams.maxmin, fbaParams.objStr)
        # Multiply by -1 for maximization
        maxmin_factor = -1. if maxmin == True else 1.
        nCols = len(lb)
        for value in lb:
            if value < 0.:
                print ("Error: Minimization of total flux requires non-negative"
                       " flux variables.")
                exit()

        # Get additional linear equality and inequality constraints
        try:
            Aeq, beq, Aineq, bineq = self.makeConstraintMatrices(matrix,
                                                                 *ParamParser.linConstraintsToVectors(fbaParams.linConstraints,
                                                                                                      reactions, nCols))
        except ValueError:
            # If any linear constraint is contradictory, return empty solution
            return 0., []

        # Prepare additional constraint: total flux < threshold
        if Aineq is None:
            Aineq, bineq = array([[1.]*nCols]), []
        else:
            # Aineq has extra line for total flux
            Aineq = vstack((Aineq, [[1.]*nCols]))
            # Convert bineq to list (allows easy addition of extra value)
            bineq = list(bineq)

        # Build linear objective function (given as coefficient vector)
        try:
            objective = ParamParser.convertObjFuncToLinVec(objStr, reactions,
                                                           nCols, maxmin)
        except Exception as strerror:
            print ("Error while trying to build coefficient vector of "
                   "linear objective function:")
            print(strerror)
            exit()

        if start_value_total is None or start_value_total < 0.:
            start_value_total = float(len(reactions))

        # Prepare left and right border for binary search
        l = start_value_total / 2.
        r = start_value_total * 1.5
        m = start_value_total

        # Use GLPK through OpenOpt
        solver = (_OOGLPK if (self.solver.lower() == _GLPK or
                              self.solver.lower() == "default") else self.solver)
        try:
            ps = LinearProblem(Aeq, beq, Aineq, bineq+[m], lb, ub, solver)
        except ValueError as strerror:
            print(strerror)
            exit()
        ps.setObjective(objective)

        try:
            m_value = ps.minimize()[0]
        except ValueError as strerror:
            if self.solver == "default":
                print("Default solver reported an error:")
            else:
                print("Solver %s reported an error:" % self.solver)
            print(strerror)
            print("Consider trying a different solver.")
            exit()
        except KeyError as strerror:
            print(strerror)
            exit()

        ps.bineq[-1] = l
        l_value, l_flux = ps.minimize()
        #print "l_flux", l_flux
        ps.bineq[-1] = r
        r_value = ps.minimize()[0]

        nSteps = 3     # Counter for calls of ps.minimize()
        print ("Step %u: l=%.10g (%.10g), r=%.10g (%.10g), m=%.10g (%.10g)" %
               (nSteps, l, l_value, r, r_value, m, m_value))

        # 1. Find right border so that f(m) = f(r)
        while abs(m_value-r_value) >= tolerance_obj:
            r *= 10.
            ps.bineq[-1] = r
            r_value = ps.minimize()[0]
            m = (l+r)/2.
            ps.bineq[-1] = m
            m_value = ps.minimize()[0]
            nSteps += 2
            print ("Step %u: l=%.10g (%.10g), r=%.10g (%.10g), m=%.10g (%.10g)"
                   % (nSteps, l, l_value, r, r_value, m, m_value))
        print("Right border found.")
        print("abs(m_value-r_value):", abs(m_value-r_value))

        # 2. Find left border so that f(l) > f(m)
        l_changed = False
        while l_flux == [] or (abs(l_value-m_value) < tolerance_obj and
                               abs(l_value) >= EPSILON):
            if not l_changed:
                l_changed = True
            if l_flux == []:
                l = (l+m)/2.
            else:
                l /= 2.
            ps.bineq[-1] = l
            l_value, l_flux = ps.minimize()
            nSteps += 1

        if l_changed:
            m = (l+r)/2.
            ps.bineq[-1] = m
            m_value = ps.minimize()[0]
            nSteps += 1
            print ("Step %u: l=%.10g (%.10g), r=%.10g (%.10g), m=%.10g (%.10g)"
                   % (nSteps, l, l_value, r, r_value, m, m_value))
        print("Left border found.")
        print("abs(m_value-l_value):", abs(m_value-l_value))

        while (abs(r-l) >= tolerance_total and
               abs(l_value-r_value) >= tolerance_obj):

            if abs(m_value-r_value) < tolerance_obj:
                r, r_value = m, m_value

            else:
                l, l_value = m, m_value

            m = (l+r)/2.
            ps.bineq[-1] = m
            m_value, solution = ps.minimize()
            nSteps += 1

            print ("Step %u: l=%.10g (%.10g), r=%.10g (%.10g), m=%.10g (%.10g)"
                   % (nSteps, l, l_value, r, r_value, m, m_value))

        if r_value < m_value:
            m = r
            ps.bineq[-1] = m
            m_value, solution = ps.minimize()
            nSteps += 1
            print ("Step %u: l=%.10g (%.10g), r=%.10g (%.10g), m=%.10g (%.10g)"
                   % (nSteps, l, l_value, r, r_value, m, m_value))

        return m, m_value * maxmin_factor, solution, nSteps

    def runOnModel(self, model, fbaParams, splitFluxes=True,
                   minimizeTotalFlux=False, rmDeadEnds=True):
        """ run flux-balance analysis on the given MetabolicModel

        Keyword arguments:

        model       -- the MetabolicModel with reactions and bounds
        fbaParams   -- optimization parameters (incl. objective function)
        splitFluxes -- if True, run split fluxes (resulting in non-negative flux
                       variables) before analysis
        minimizeTotalFlux -- if True, find a solution with minimal total absolute flux
        rmDeadEnds   -- if True, remove all reactions with dead ends before
                        analysis (faster and gives an optimal solution, as well)

        Returns:
        obj_value, solution, dimReduced

        obj_value   -- optimal value of objective function
        solution    -- a solution where the objective function assumes obj_value
        dimReduced  -- pair (nRows, nColumns) with dimensions of reduced matrix
        """
        maxmin = fbaParams.maxmin
        objStr = fbaParams.objStr

        if isinstance(maxmin, str):
            maxmin = {"max": True, "min": False}[maxmin.lower()]

        if rmDeadEnds:
            deadReactions = model.findDeadEnds(True)[1]
            modelRed = model.getSubModelByExcludeList(deadReactions)
            cbz = model.canBeZero(deadReactions)
            nonZeroDeadEnds = [deadReactions[i] for i in
                               range(len(deadReactions)) if not cbz[i]]
            if nonZeroDeadEnds:
                print ("The following blocked reactions are constrained to a "
                       "non-zero flux:\n  " + "\n  ".join(nonZeroDeadEnds) +
                       "\nThe problem is infeasible.")
                return (0., MetabolicFlux(),
                        array(modelRed.getStoichiometricMatrix()).shape)
        else:
            modelRed = model

        matrix = array(modelRed.getStoichiometricMatrix())
        dimReduced = matrix.shape
        reaction_names = modelRed.getReactionNames()
        reactions = modelRed.reactionDict
        lb, ub = modelRed.getBounds()

        if splitFluxes:
            matrixSplit, reactionsSplit, lbSplit, ubSplit = \
                self.splitFluxes(matrix, reaction_names, lb, ub)
            try:
                ParamParser.convertObjFuncToLinVec(objStr, reactionsSplit,
                                                   len(lbSplit), maxmin)
            except KeyError as s:
                if deadReactions:
                    if self.verb:
                        print ("Error while trying to construct the objective "
                               "function vector:\n%s\nThis may be due to removal "
                               "of nonfunctional reactions." % s)
                    return 0., MetabolicFlux(), dimReduced

            obj_value, solutionSplit = self.run(reactionsSplit, matrixSplit,
                                                lbSplit, ubSplit, fbaParams)
            solution = []
            try:
                if solutionSplit == []:
                    solution = []
                else:
                    if minimizeTotalFlux:
                        # Optimize solution by limiting total flux
                        limit, obj_value, solutionSplit, nSteps = \
                            self.runWithTotalFluxMinimization(reactionsSplit,
                                                              matrixSplit, lbSplit, ubSplit, fbaParams,
                                                              sum(solutionSplit), 1e-8, 1e-12)
                        print ("Found optimal flux limit of %.10g in %u steps."
                               % (limit, nSteps))

                    solution = array(FbAnalyzer.rejoinFluxes(solutionSplit,
                                                             reactionsSplit, reactions))
            except NameError as strerror:
                print(strerror)

        else:
            obj_value, solution = self.run(
                reactions, matrix, lb, ub, fbaParams)
        flux = MetabolicFlux(modelRed, solution)

        # Add removed reactions with flux 0. and original bounds to solution
        if len(modelRed) != len(model) and len(solution) != 0:
            reactionsRed = set(reaction_names)
            for rea in model:
                if rea.name not in reactionsRed:
                    flux.fluxDict[rea.name] = 0.
                    flux.boundsDict[rea.name] = (rea.lb, rea.ub)

        return obj_value, flux, dimReduced


def main():
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "Flux balance analysis\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile",
                      help="perform Flux Balance Analysis on the network given "
                           "by the reaction FILE", metavar="FILE")
    parser.add_option("-p", "--parameters", dest="paramFile",
                      help="use the given scenario FILE for Flux Balance "
                           "Analysis", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile",
                      help="write output of Flux Balance Analysis to FILE",
                      metavar="FILE")
    parser.add_option("-m", "--minimize-total", action="store_true",
                      dest="minimizeFlux", help="secondarily minimize total "
                      "flux")
    parser.add_option("-l", "--use-full-matrix", action="store_true",
                      dest="useFullMatrix", help="use full matrix (disable "
                      "removal of dead ends and nonfunctional reactions)")
    parser.set_defaults(minimizeFlux=False, useFullMatrix=False)

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-p")
    parser.check_required("-o")

    # 2. Parse reaction file

    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 3. Parse scenario file

    model_messages = []
    pparser = ParamParser()
    try:
        # Parse file, get maxmin, name of objective function, and solver name
        maxmin, objStr, solver, numIter, lb, ub = pparser.parse(
            options.paramFile)
        fbaParams = FbaParam(solver, maxmin, objStr, numIter)
        fbaParams.setLinConstraints(pparser.lin_constraints)
        # Set flux bounds in model
        model.setFiniteBounds(lb, ub, True, model_messages)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in scenario file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except ValueError as strerror:
        print(strerror)
        exit()

    # 4. Perform flux balance analysis on the metabolic network

    if solver == "":
        solver = "default"
    fba = FbAnalyzer(solver)
    obj_value, solution, ndim = fba.runOnModel(model, fbaParams, True,
                                               options.minimizeFlux, not options.useFullMatrix)

    # Show warning and info messages of parsers
    msgs = (rparser.getMessages() + pparser.getMessages() +
            [x[1] for x in model_messages])
    if msgs:
        print('\n'+'\n'.join(msgs))
    print ("Info: The reduced network has %u reactions and %u metabolites." %
           ndim[1::-1])

    # 5. Write output to file
    if len(solution) != 0:
        print("\nValue of objective function at solution:", obj_value)
        print("Total absolute flux:", end=' ')
        print(abs(array(list(solution.fluxDict.values()))).sum())
        try:
            solution.writeToFile(options.outputFile)
        except IOError as strerror:
            print ("Unable to write to file %s:" %
                   os.path.basename(options.outputFile))
            print(strerror)
            exit()
    else:
        print("No output written.")
        exit()


if __name__ == "__main__":
    main()



#!/usr/bin/env python

""" metano Model Assertion Monitor

This script reads in a metabolic model from a reaction file and a scenario file,
performs FBA, FVA, and split-ratio analysis and checks the results against a
list of assertions. All assertions that do not hold are reported.

It must never be used in a non-safe environment (such as in a server-based
application), because eval() is used to evaluate the assertions.

Allowed symbols are as follows:
<reaction_name>      -- flux through reaction in FBA solution
min(<reaction_name>) -- FVA minimum of flux through reaction
max(<reaction_name>) -- FVA maximum        - '' -
<metabolite_name>    -- total flux going through metabolite (computed as influx)
inratio(<metabolite_name>, <reaction_name>)
                     -- fraction of metabolite produced via reaction
outratio(<metabolite_name>, <reaction_name>)
                     -- fraction of metabolite consumed via reaction


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from builtins import zip
from builtins import range
from builtins import object
from metano.defines import FbaParam, COPYRIGHT_VERSION_STRING
from metano.fba import OptionParser, FbAnalyzer
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.metabolicmodel import MetabolicModel
from metano.metabolicflux import MetabolicFlux
from metano.fva import FvAnalyzer
import os
import re


class ModelWatcher(object):
    """ Class for plausibility analysis of metabolic models
    """

    commentSign = '#'

    def __init__(self, model=None, fvaTolerance=.95):
        """ initialize watcher class

        Keyword arguments:

        model        -- MetabolicModel to be checked
        fvaTolerance -- objective function threshold for FVA
        """
        self.setModel(model)
        self.clearAssertions()
        self.fvaTolerance = fvaTolerance

    def clearAssertions(self):
        self.assertions, self.assertions_orig = [], []

    def setModel(self, model):
        """ set the MetabolicModel and reset flux and related variables
        """
        self.model = model
        self.flux = MetabolicFlux()
        self.metFlux, self.inRatios, self.outRatios = {}, {}, {}
        self.minFlux, self.maxFlux = {}, {}

    def readAssertionsFromFile(self, filename):
        """ parse the assertion file identified by filename
            -- wrapper for readAssertionsFromFileHandle()
        """
        with open(filename) as f:
            return self.readAssertionsFromFileHandle(f)

    def readAssertionsFromFileHandle(self, f):
        """ read the assertion file given as a file object

        Modified member variables:

        assertions_orig -- list of assertions in original format
        assertions      -- list of Python statements to be evaluated

        Comments and empty (or all-space) lines are removed
        """
        s = f.read()

        # Remove comments - everything from comment sign ('#') to end of line
        commentPos = 0
        while commentPos >= 0:
            commentPos = s.find(self.commentSign, commentPos)
            if commentPos >= 0:
                eolPos = s.find('\n', commentPos)
                if eolPos < 0:
                    # Last line - delete everything starting from comment sign
                    s = s[:commentPos]
                    commentPos = -1
                else:
                    # Excise comment
                    s = s[:commentPos]+s[eolPos:]

        if s.find(':') >= 0:
            raise SyntaxError('Illegal character found (:)')

        self.assertions_orig = s.split('\n')

        # Replace reaction and metabolite names with unique non-matching IDs
        # including a type specifier (r or m)

        idDict = dict(list(zip(list(self.model.reactionDict.keys()), "r"*len(self.model))) +
                      list(zip(list(self.model.metaboliteDict.keys()),
                               "m"*len(self.model.metaboliteDict))))

        # Find non-overlapping longest matches (greedy) of reaction and
        # metabolite names and replace with unique non-matching IDs
        counter = 0
        replaceDict, replaceDictRev = {}, {}
        for ID in sorted(list(idDict.keys()), key=len, reverse=True):
            pos = 0
            while pos >= 0:
                pos = s.find(ID, pos)
                if pos >= 0:
                    if ID in replaceDict:
                        replaceId = replaceDict[ID]
                    else:
                        # Generate new ID
                        replaceId = ":%u%c:" % (counter, idDict[ID])
                        replaceDict[ID] = replaceId
                        replaceDictRev[replaceId] = ID
                        counter += 1

                    s = s[:pos]+replaceId+s[pos+len(ID):]
                    pos = pos+len(replaceId)+1

        reaString = r"(:\d+r:)"
        metString = r"(:\d+m:)"
        inratioPattern = re.compile(r"inratio\s*\(\s*"+metString+r"\s*,\s*" +
                                    reaString+r"\s*\)")
        outratioPattern = re.compile(r"outratio\s*\(\s*"+metString+r"\s*,\s*" +
                                     reaString+r"\s*\)")
        minPattern = re.compile(r"min\s*\(\s*"+reaString+r"\s*\)")
        maxPattern = re.compile(r"max\s*\(\s*"+reaString+r"\s*\)")
        reaPattern = re.compile(reaString)
        metPattern = re.compile(metString)

        # Replace inratio/outratio expressions with references to self.inRatios/
        #                                                         self.outRatios
        s = inratioPattern.sub(lambda x: "self.inRatios[%r].get(%r, 0.)" %
                               (replaceDictRev[x.group(1)],
                                replaceDictRev[x.group(2)]), s)
        s = outratioPattern.sub(lambda x: "self.outRatios[%r].get(%r, 0.)" %
                                (replaceDictRev[x.group(1)],
                                 replaceDictRev[x.group(2)]), s)

        # Replace min/max expressions with references to self.minFlux/
        #                                                self.maxFlux
        s = minPattern.sub(lambda x: "self.minFlux[%r]" %
                           replaceDictRev[x.group(1)], s)
        s = maxPattern.sub(lambda x: "self.maxFlux[%r]" %
                           replaceDictRev[x.group(1)], s)

        # Replace reaction identifiers with references to self.flux
        s = reaPattern.sub(lambda x: "self.flux[%r]" %
                           replaceDictRev[x.group(0)], s)

        # Replace metabolite identifiers with references to self.metFlux
        s = metPattern.sub(lambda x: "self.metFlux[%r]" %
                           replaceDictRev[x.group(0)], s)

        self.assertions = s.split('\n')

        # Remove blank lines (cannot be evaluated)
        for i in range(len(self.assertions)-1, -1, -1):
            if self.assertions[i] == "" or self.assertions[i].isspace():
                del self.assertions[i]
        for i in range(len(self.assertions_orig)-1, -1, -1):
            if (self.assertions_orig[i] == "" or
                    self.assertions_orig[i].isspace()):
                del self.assertions_orig[i]

    def checkSyntax(self):
        """ check assertions for syntax errors without actually performing any
            potentially time-consuming analyses

        This function prints a message to the console for every assertion that
        contains a syntax error.

        Returns: True if all assertions can be evaluated, False if error occurs
        """
        # Backup member variables and re-initialize with dummy values to make
        # sure that all references exist
        bakflux, bakMetFlux = self.flux, self.metFlux
        bakInRatios, bakOutRatios = self.inRatios, self.outRatios
        bakMinFlux, bakMaxFlux = self.minFlux, self.maxFlux

        self.flux = MetabolicFlux(self.model, [0.]*len(self.model))
        self.metFlux = dict(list(zip(self.model.metabolites,
                                     [0.]*len(self.model.metabolites))))
        self.inRatios = dict(list(zip(self.model.metabolites,
                                      [{}]*len(self.model.metabolites))))
        self.outRatios = dict(list(zip(self.model.metabolites,
                                       [{}]*len(self.model.metabolites))))
        self.minFlux = dict(list(zip(list(self.model.reactionDict.keys()),
                                     [0.]*len(self.model))))
        self.maxFlux = dict(list(zip(list(self.model.reactionDict.keys()),
                                     [0.]*len(self.model))))
        excepted = []
        for i in range(len(self.assertions)):
            try:
                eval(self.assertions[i])
            except (SyntaxError, NameError):
                excepted.append(repr(self.assertions_orig[i]))

        # Restore member variables to original values
        self.flux, self.metFlux = bakflux, bakMetFlux
        self.inRatios, self.outRatios = bakInRatios, bakOutRatios
        self.minFlux, self.maxFlux = bakMinFlux, bakMaxFlux

        if excepted:
            print ("The following assertions cannot be evaluated due to errors"
                   ":\n  " + "\n  ".join(excepted))
#  The following was removed because the Python error messages are misleading.
#            for a in excepted:
#                print "\n  %r\n  %s: %s\n" % (a[0], a[1], a[2])
            return False
        return True

    def run(self, fbaParams, syntaxOk=True, printExcepted=False):
        """ check assertions against results of FBA, FVA, & split-ratio analysis

        This function performs FBA, split-ratio analysis, and FVA and evaluates
        the assertions. It enumerates all assertions that do not hold.

        Keyword arguments:
        fbaParams     -- objective function and solver for FBA
        syntaxOk      -- if False, previously performed syntax check has failed
        printExcepted -- if False, syntax errors in assertions are not reported
        """

        # 1. Perform flux balance analysis on the metabolic network

        fba = FbAnalyzer(fbaParams.solver)
        self.flux = fba.runOnModel(self.model, fbaParams)[1]
        if len(self.flux) == 0:
            print ("No FBA solution was obtained. The optimization problem is "
                   "either unbounded or infeasible.")
            return

        # 2. Compute metabolite fluxes and split ratios

        splitRatios = self.flux.computeAllSplitRatios(self.model)
        self.metFlux, self.outRatios, self.inRatios = {}, {}, {}
        for met in splitRatios:
            outRatios, inRatios = splitRatios[met]
            self.metFlux[met] = sum(inRatios[rea][1] for rea in inRatios)
            self.outRatios[met] = dict((rea, outRatios[rea][0])
                                       for rea in outRatios)
            self.inRatios[met] = dict((rea, inRatios[rea][0])
                                      for rea in inRatios)

        # 3. Perform flux variability analysis on the model

        fva = FvAnalyzer("default")  # always use GLPK (=> fastFVA)
        minmax = fva.runOnModel(self.model, fbaParams, self.fvaTolerance,
                                self.flux)[0]

        self.minFlux, self.maxFlux = {}, {}
        for i in range(len(self.model)):
            self.minFlux[self.model.reactions[i].name], self.maxFlux[
                self.model.reactions[i].name] = minmax[i]

        # 4. Check assertions

        failed = []
        excepted = []
        for i in range(len(self.assertions)):
            try:
                if not eval(self.assertions[i]):
                    failed.append(repr(self.assertions_orig[i]))
            except (SyntaxError, NameError):
                excepted.append(repr(self.assertions_orig[i]))

        if printExcepted and excepted:
            print ("The following assertions could not be evaluated due to "
                   "errors:\n  " + "\n  ".join(excepted))
        if failed:
            print("The following assertions failed:\n  " + "\n  ".join(failed))
        else:
            if not syntaxOk or (printExcepted and excepted):
                print("All other assertions hold.")
            else:
                print("All assertions hold.")


def main():
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "Model assertion watcher\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile", help="read "
                      "metabolic model from the given reaction FILE",
                      metavar="FILE")
    parser.add_option("-p", "--parameters", dest="paramFile", help="use the "
                      "given scenario FILE for model analysis ", metavar="FILE")
    parser.add_option("-a", "--assertions", dest="assertionFile", help="read "
                      "assertions from FILE",  metavar="FILE")
    parser.add_option("-t", "--tolerance", dest="tolerance", type="float",
                      help="FVA: tolerance for objective function (0 < VALUE < "
                      "1; default: .95)", metavar="VALUE")
    parser.set_defaults(tolerance=.95)

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-p")
    parser.check_required("-a")

    if options.tolerance <= 0. or options.tolerance > 1.:
        print("Error: Tolerance must be in interval (0, 1]")
        exit()

    # 2. Create MetabolicModel from reaction file

    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 3. Parse scenario file

    model_messages = []
    pparser = ParamParser()
    try:
        # Parse file, get maxmin, name of objective function, and solver name
        maxmin, objStr, solver, numIter, lb, ub = \
            pparser.parse(options.paramFile)
        if solver == "":
            solver = "default"
        fbaParams = FbaParam(solver, maxmin, objStr, numIter)
        # Set flux bounds in model
        model.setFiniteBounds(lb, ub, True, model_messages)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in scenario file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except ValueError as strerror:
        print(strerror)
        exit()

    # 4. Create ModelWatcher object and parse assertion file

    maw = ModelWatcher(model, options.tolerance)
    try:
        maw.readAssertionsFromFile(options.assertionFile)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.assertionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in assertion file %s:" %
               os.path.basename(options.assertionFile))
        print(strerror)
        exit()

    # 5. First perform syntax check, then evaluate assertions

    maw.run(fbaParams, maw.checkSyntax())


if __name__ == "__main__":
    main()



"""
This module defines the class MetabolicFlux, which represents a vector of fluxes
through the reactions of a MetabolicModel.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import division

from builtins import zip
from builtins import map
from builtins import object
from past.utils import old_div
import csv
from numpy import inf
from metano.defines import typename, addToDictPlus, padNumber
from metano.metabolicmodel import MetabolicModel


class MetabolicFlux(object):
    """ Represents a list of fluxes through a metabolic network. The list is
        stored internally as a dict {reaction name : flux value}.

    Member variables:
    fluxDict   -- dict {name : flux value}
    boundsDict -- dict {name : (lb, ub)} of lower and upper bounds for each flux
    """

    def __init__(self, reactions=None, fluxVec=None, bounds=None):
        """ initialize a metabolic flux instance from one of the following:
            a) (nothing)
            b) dictionary {reaction name : flux value}
            c) MetabolicModel, list of values
            d) list of names, list of values

        Keyword arguments:
        reactions -- dictionary, MetabolicModel, or list of names (see above)
        fluxVec   -- list of flux values (also see above)
        bounds    -- dictionary {reaction name : (lower bound, upper bound)}
        """
        self.clear()
        if not reactions:
            return
        if isinstance(reactions, dict):
            self.fluxDict = reactions
        else:
            try:
                if isinstance(reactions, MetabolicModel):
                    self.setByModel(reactions, fluxVec)
                else:
                    self.setByList(reactions, fluxVec)
            except TypeError:
                raise TypeError("Expected dict or two lists or MetabolicModel "
                                "and list; got (%s, %s)" % (typename(reactions),
                                                            typename(fluxVec)))
        if bounds:
            self.boundsDict = bounds

    def __len__(self):
        return len(self.fluxDict)

    def __getitem__(self, key):
        return self.fluxDict[key]

    def __setitem__(self, key, value):
        self.fluxDict[key] = value

    def __delitem__(self, key):
        del self.fluxDict[key]

    def __iter__(self):
        return iter(self.fluxDict)

    def __repr__(self):
        if not self.fluxDict and not self.boundsDict:
            return "MetabolicFlux()"
        return "MetabolicFlux(%r, bounds=%r)" % (self.fluxDict, self.boundsDict)

    def __str__(self):
        if not self.fluxDict:
            return "MetabolicFlux()"  # Representation of empty object
        s = ""
        # Get length of longest reaction name (for formatting)
        maxlen = len(max(self.fluxDict, key=len))
        nFluxes = len(self.fluxDict)
        count = 0
        for name in sorted(self.fluxDict):
            try:
                lb, ub = self.boundsDict[name]
            except KeyError:
                lb, ub = -inf, inf
            s += "%s : %g (%g, %g)" % (name.ljust(maxlen), self.fluxDict[name],
                                       lb, ub)
            count += 1
            if count < nFluxes:
                s += "\n"
        return s

    def clear(self):
        self.fluxDict = {}
        self.boundsDict = {}

    def clearBounds(self):
        self.boundsDict = {}

    def setBounds(self, name, lb, ub):
        self.boundsDict[name] = lb, ub

    def setToZero(self):
        """ sets all fluxes to zero
        """
        for name in self.fluxDict:
            self.fluxDict[name] = 0.

    def setByModel(self, model, fluxVec):
        """ assign the fluxes given by fluxVec to the reactions in the given
            model - fluxVec[i] is assigned to the i-th reaction of the model;
            bounds are also set from the model

        Keyword arguments:

        model   -- a MetabolicModel
        fluxVec -- vector of fluxes, length must be number of reactions in the
                   model

        This function does not clear self.fluxDict beforehand.
        """
        if len(fluxVec) == 0:
            return  # Nothing to do (e.g. if FBA solution is empty)
        if len(model) != len(fluxVec):
            raise ValueError("Length of flux vector and number of reactions "
                             "in the model don't agree")
        for (name, flux) in zip(model.getReactionNames(), fluxVec):
            self.fluxDict[name] = flux
        lb, ub = model.getBounds()
        for name, bounds in zip(model.getReactionNames(), list(zip(lb, ub))):
            self.boundsDict[name] = bounds

    def setByDict(self, d, fluxVec):
        """ set the fluxes from the given vector with the order from the given
            dictionary

        Keyword arguments:

        d       -- dictionary {name : index} for ordering fluxVec
        fluxVec -- vector of fluxes, length must be number of reactions in the
                   model

        This function does not clear self.fluxDict beforehand.
        """
        if len(d) != len(fluxVec):
            raise ValueError(
                "Length of flux vector and dictionary don't agree")
        for name in d:
            self.fluxDict[name] = fluxVec[d[name]]

    def setByList(self, names, fluxVec):
        """ set the fluxes from the given vector with the order from the given
            dictionary

        Keyword arguments:

        d       -- dictionary {name : index} for ordering fluxVec
        fluxVec -- vector of fluxes, length must be number of reactions in the
                   model

        This function does not clear self.fluxDict beforehand.
        """
        if len(names) != len(fluxVec):
            raise ValueError("Length of name list and flux vector don't agree")
        self.fluxDict = dict(list(zip(names, fluxVec)))

    def getVecOrderedByModel(self, model):
        """ return a flux vector ordered like the reactions in the model

        Fluxes through reactions not present in the model are ignored.
        """
        fluxVec = []
        for name in model.getReactionNames():
            if name in self.fluxDict:
                fluxVec.append(self.fluxDict[name])
            else:
                fluxVec.append(None)  # Reaction has no flux value
        return fluxVec

    def getVecOrderedByDict(self, d):
        """ return a flux vector with the order from the given dictionary

        Keyword arguments:

        d       -- dictionary {name : index} for ordering fluxVec
        """
        fluxVec = [0.]*len(d)
        for name in d:
            if name in self.fluxDict:
                fluxVec[d[name]] = self.fluxDict[name]
            else:
                fluxVec[d[name]] = None  # Reaction has no flux value
        return fluxVec

    def hasSameReactions(self, reactions):
        """ verify that the MetabolicFlux object has the same reactions as the
            given 'reactions' object, which can be a MetabolicModel, a dict,
            a sequence type or a MetabolicFlux object
        """
        # Get set of reactions (to eliminate possible duplicates)
        if isinstance(reactions, MetabolicModel):
            reactionSet = set(reactions.getReactionNames())
        else:
            reactionSet = set(reactions)

        if len(self.fluxDict) != len(reactionSet):
            return False
        for rea in reactionSet:
            if rea not in self.fluxDict:
                return False
        return True

    def absDiff(self, other):
        """ return the absolute difference to the other MetabolicFlux as
            MetabolicFlux object; if a reaction is only contained in self or
            other, the flux is treated as if it were zero in the other object
        """
        diffDict = {}
        for rea in self.fluxDict:
            if rea in other.fluxDict:
                diffDict[rea] = abs(self.fluxDict[rea]-other.fluxDict[rea])
            else:
                diffDict[rea] = abs(self.fluxDict[rea])

        for rea in other.fluxDict:
            if rea not in self.fluxDict:
                diffDict[rea] = abs(other.fluxDict[rea])
        return MetabolicFlux(diffDict)

    def strictAbsDiff(self, other):
        """ return the absolute difference to the other MetabolicFlux as
            MetabolicFlux object - both must have exactly the same reactions
        """
        diffDict = {}
        if len(self.fluxDict) != len(other.fluxDict):
            raise ValueError("Lengths don't agree")
        for name in self.fluxDict:
            if name not in other.fluxDict:
                raise ValueError("Reaction '%s' not found in other "
                                 "MetabolicFlux object" % name)
            diffDict[name] = abs(self.fluxDict[name]-other.fluxDict[name])
        return MetabolicFlux(diffDict)

    def sqDiff(self, other):
        """ return the squared difference to the other MetabolicFlux as
            MetabolicFlux object; if a reaction is only contained in self or
            other, the flux is treated as if it were zero in the other object
        """
        diffDict = {}
        for rea in self.fluxDict:
            if rea in other.fluxDict:
                diff = self.fluxDict[rea]-other.fluxDict[rea]
            else:
                diff = self.fluxDict[rea]
            diffDict[rea] = diff*diff

        for rea in other.fluxDict:
            if rea not in self.fluxDict:
                diff = other.fluxDict[rea]
                diffDict[rea] = diff*diff
        return MetabolicFlux(diffDict)

    def strictSqDiff(self, other):
        """ return the squared difference to the other MetabolicFlux as
            MetabolicFlux object - both must have exactly the same reactions
        """
        diffDict = {}
        if len(self.fluxDict) != len(other.fluxDict):
            raise ValueError("Lengths don't agree")
        for name in self.fluxDict:
            if name not in other.fluxDict:
                raise ValueError("Reaction '%s' not found in other "
                                 "MetabolicFlux object" % name)
            diff = self.fluxDict[name]-other.fluxDict[name]
            diffDict[name] = diff*diff
        return MetabolicFlux(diffDict)

    def computeSplitRatios(self, metabolite, model, cutoff=0.,
                           cutoffIsAbsolute=False, listAll=False):
        """ compute split ratios for the fluxes entering and leaving the given
            metabolite node

        Keyword arguments:

        metabolite       -- name of the metabolite
        model            -- the underlying MetabolicModel
        cutoff           -- cutoff value below which fluxes are considered
                            absent, must be >= 0
        cutoffIsAbsolute -- True : cutoff is in absolute flux units,
                            False: cutoff is relative (between 0 and 1; default)
        listAll          -- if True, ignore cutoff and list all fluxes (even 0)

        Returns:
        outRatios, inRatios

        outRatios        -- dict { reaction : (ratio, flux) } for all outgoing
                            fluxes, i.e. the fluxes consuming the metabolite
        inRatios         -- dict { reaction : (ratio, flux) } for all incoming
                            fluxes, i.e. the fluxes producing the metabolite
        """
        posDict, negDict = {}, {}
        for rea in model:
            for coef, met in rea:
                if met == metabolite:
                    # Compute flux through metabolite from current reaction
                    metFlux = self.fluxDict[rea.name]*coef
                    if listAll:
                        # Zero fluxes are arbitrarily counted as outgoing
                        if metFlux > 0.:
                            addToDictPlus(posDict, rea.name, metFlux)
                        else:
                            addToDictPlus(negDict, rea.name, -metFlux)
                    if cutoffIsAbsolute:
                        if metFlux > cutoff:
                            addToDictPlus(posDict, rea.name, metFlux)
                        elif metFlux < -cutoff:
                            addToDictPlus(negDict, rea.name, -metFlux)
                    else:
                        # Fluxes of exactly zero can be discarded already here
                        if metFlux > 0.:
                            addToDictPlus(posDict, rea.name, metFlux)
                        elif metFlux < 0.:
                            addToDictPlus(negDict, rea.name, -metFlux)

        dontCheck = listAll or cutoffIsAbsolute

        posSum = sum(posDict.values())
        reactions = list(posDict.keys())
        for rea in reactions:
            value = posDict[rea]
            ratio = 0. if posSum == 0. else old_div(value, posSum)
            if dontCheck or ratio > cutoff:
                posDict[rea] = ratio, value
            else:
                del posDict[rea]

        negSum = sum(negDict.values())
        reactions = list(negDict.keys())
        for rea in reactions:
            value = negDict[rea]
            ratio = 0. if negSum == 0. else old_div(value, negSum)
            if dontCheck or ratio > cutoff:
                negDict[rea] = ratio, value
            else:
                del negDict[rea]

        return negDict, posDict

    def computeAllSplitRatios(self, model, cutoff=0., cutoffIsAbsolute=False,
                              listAll=False):
        """ compute metabolite fluxes and split ratios for the fluxes entering
            and leaving each metabolite node in the given model

        Keyword arguments:

        model            -- the underlying MetabolicModel
        cutoff           -- cutoff value below which fluxes are considered
                            absent, must be >= 0
        cutoffIsAbsolute -- True : cutoff is in absolute flux units,
                            False: cutoff is relative (between 0 and 1; default)
        listAll          -- if True, ignore cutoff and list all fluxes (even 0)

        Returns:
        splitRatios      -- dict { metabolite : (outRatios, inRatios) } with
                            outRatios and inRatios as returned by
                            computeSplitRatios()
        """
        splitRatios = dict((met, ({}, {}))
                           for met in model.getMetaboliteNames())
        for rea in model:
            for coef, met in rea:
                # Compute flux through metabolite 'met' from reaction 'rea'
                metFlux = self.fluxDict[rea.name]*coef
                if listAll:
                    # Zero fluxes are arbitrarily counted as outgoing
                    if metFlux > 0.:
                        addToDictPlus(splitRatios[met][1], rea.name, metFlux)
                    else:
                        addToDictPlus(splitRatios[met][0], rea.name, -metFlux)
                elif cutoffIsAbsolute:
                    if metFlux > cutoff:
                        addToDictPlus(splitRatios[met][1], rea.name, metFlux)
                    elif metFlux < -cutoff:
                        addToDictPlus(splitRatios[met][0], rea.name, -metFlux)
                else:
                    # Fluxes of exactly zero can be discarded already here
                    if metFlux > 0.:
                        addToDictPlus(splitRatios[met][1], rea.name, metFlux)
                    elif metFlux < 0.:
                        addToDictPlus(splitRatios[met][0], rea.name, -metFlux)

        dontCheck = listAll or cutoffIsAbsolute

        for met in splitRatios:
            negDict, posDict = splitRatios[met]
            posSum = sum(posDict.values())
            reactions = list(posDict.keys())
            for rea in reactions:
                value = posDict[rea]
                ratio = 0. if posSum == 0. else old_div(value, posSum)
                if dontCheck or ratio > cutoff:
                    posDict[rea] = ratio, value
                else:
                    del posDict[rea]

            negSum = sum(negDict.values())
            reactions = list(negDict.keys())
            for rea in reactions:
                value = negDict[rea]
                ratio = 0. if negSum == 0. else old_div(value, negSum)
                if dontCheck or ratio > cutoff:
                    negDict[rea] = ratio, value
                else:
                    del negDict[rea]

        return splitRatios

    # ------ File I/O ----------------------------------------------------------

    def writeToFile(self, filename):
        """ write to the given file
        """
        with open(filename, 'w') as f:
            if filename.split(".")[-1] == "csv":
                self.writeToCSVHandle(f)
            else:
                self.writeToFileHandle(f)

    def writeToCSVHandle(self, f):
        """ write to the CSV file given by file handle f (must be open for writing)
        """
        if not self.fluxDict:
            return  # Nothing to do

        csvwriter = csv.writer(
            f, delimiter=";", quotechar='"', quoting=csv.QUOTE_MINIMAL)
        # Write Header to CSV file
        csvwriter.writerow(["NAME", "FLUX", "LB", "UB"])
        result = []
        # Construct output table and get widths of table columns (length of
        # longest occurring string)
        for rea in self.fluxDict:
            try:
                lb, ub = self.boundsDict[rea]
            except KeyError:
                lb, ub = -inf, inf
            # update python3
            result.append([rea.name if type(rea) != str else rea,
                           self.fluxDict[rea], lb, ub])

        for row in sorted(result):
            csvwriter.writerow(row)

    def writeToFileHandle(self, f):
        """ write to the file given by file handle f (must be open for writing)
        """
        if not self.fluxDict:
            return  # Nothing to do

        # Construct output table and get widths of table columns (length of
        # longest occurring string)
        maxlenName, maxlenFlux, maxlenLb, maxlenUb = 0, 0, 0, 0
        outputTable = {}  # dict {name : (flux, lb, ub)} - all as strings
        for rea in self.fluxDict:
            try:
                lb, ub = self.boundsDict[rea]
            except KeyError:
                lb, ub = -inf, inf

            # Pad non-negative numbers with space
            outputTable[rea] = tuple(map(padNumber, list(map(repr,
                                                             (self.fluxDict[rea], lb, ub)))))
            lenName, lenFlux, lenLb, lenUb = list(
                map(len, (rea,)+outputTable[rea]))

            if lenName > maxlenName:
                maxlenName = lenName
            if lenFlux > maxlenFlux:
                maxlenFlux = lenFlux
            if lenLb > maxlenLb:
                maxlenLb = lenLb
            if lenUb > maxlenUb:
                maxlenUb = lenUb

        # Write table head
        f.write("NAME".ljust(maxlenName)+"   "+"FLUX".center(maxlenFlux)+" " +
                "LB  ".rjust(maxlenLb)+" "+"UB  ".rjust(maxlenUb)+"\n")
        # Write result for the flux through every reaction
        for rea in sorted(self.fluxDict):
            flux, lb, ub = outputTable[rea]
            # update python3
            if type(rea) != str:
                rea = rea.name
            f.write(rea.ljust(maxlenName) + " : " + flux.rjust(maxlenFlux) +
                    " " + lb.rjust(maxlenLb) + " " + ub.rjust(maxlenUb) + "\n")

    def readFromFile(self, filename):
        """ parse the given file
        """
        with open(filename, 'r') as f:
            if filename.split(".")[-1] == "csv":
                self.readFromCSVHandle(f)
            else:
                self.readFromFileHandle(f)

    def readFromCSVHandle(self, f):
        """ parse the file given by file handle f (must be open for reading)
        """
        csvreader = csv.reader(f, delimiter=";", quotechar='"')

        reactions = set()
        line_no = 0
        for line in csvreader:
            line_no += 1
            if line[0] == "NAME":
                continue

            try:
                self.fluxDict[line[0]] = float(line[1])
                self.boundsDict[line[0]] = line[2], line[3]
            except IndexError:
                raise SyntaxError("Syntax error in line %u:\nLine must "
                                  "contain exactly four values (name, flux, "
                                  "lb, ub)." % line_no)
            except ValueError:
                raise SyntaxError("Syntax error in line %u: Invalid "
                                  "floating point value." % line_no)

    def readFromFileHandle(self, f):
        """ parse the file given by file handle f (must be open for reading)
        """
        reactions = set()
        line_no = 0
        for line in f:
            line_no += 1
            if (line.lstrip().upper().startswith("NAME") or line == "" or
                    line.isspace()):
                continue

            try:
                rea, values_str = list(map(str.rstrip, line.split(":")))
            except ValueError:
                raise SyntaxError("Syntax error in line %u:\nLine must "
                                  "contain exactly one colon (':')." %
                                  line_no)
            if rea in reactions:
                raise SyntaxError("Syntax error in line %u: Duplicate "
                                  "reaction." % line_no)
            reactions.add(rea)

            try:
                values = list(map(float, values_str.split(None, 3)[:3]))
            except ValueError:
                raise SyntaxError("Syntax error in line %u: Invalid "
                                  "floating point value." % line_no)
            try:
                self.fluxDict[rea] = values[0]
                self.boundsDict[rea] = values[1], values[2]
            except IndexError:
                raise SyntaxError("Syntax error in line %u:\nLine must "
                                  "contain exactly three values (flux, lb, "
                                  "ub)." % line_no)



"""
This module defines constants for structural elements of SBML files.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

Alexander Krause contributed to this file.

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""

from builtins import object
class SbmlDef(object):
    """ Class for SBML structure elements
        (with extensions for flux balance models)
    """
    # Ubiquitous attributes
    ID = "id"
    NAME = "name"
    NOTES = "notes"
    # "html"/"body" and nested "p"/"html:p" nodes in "notes" are ignored

    # "sbml" nodes and attributes
    SBML = "sbml"
    LEVEL = "level"
    VERSION = "version"
    XMLNS = "xmlns"

    # "model" and child nodes
    MODEL = "model"
    LISTOFUNITDEFINITIONS = "listOfUnitDefinitions"
    LISTOFCOMPARTMENTS = "listOfCompartments"
    LISTOFSPECIES = "listOfSpecies"
    LISTOFREACTIONS = "listOfReactions"

    # listOfUnitDefinitions
    UNITDEFINITION = "unitDefinition"
    LISTOFUNITS = "listOfUnits"
    UNIT = "unit"
    KIND = "kind"
    SCALE = "scale"
    EXPONENT = "exponent"
    MULTIPLIER = "multiplier"

    # listOfCompartments
    COMPARTMENT = "compartment"
    COMPARTMENTOUTSIDE = "outside"

    # listOfSpecies
    SPECIES = "species"
    # "charge" is ignored (and obsolete)
    BOUNDARYCONDITION = "boundaryCondition"
    CONSTANT = "constant"  # can also appear as attribute of speciesReference and compartment
    HASONLYSUBSTANCEUNITS = "hasOnlySubstanceUnits"

    # listOfReactions
    REACTION = "reaction"
    REVERSIBLE = "reversible"
    FAST = "fast"
    LISTOFREACTANTS = "listOfReactants"
    LISTOFPRODUCTS = "listOfProducts"
    KINETICLAW = "kineticLaw"
    MATH = "math"
    CI = "ci"
    CN = "cn"
    LISTOFPARAMETERS = "listOfLocalParameters"

    # listOfReactants / listOfProducts
    SPECIESREFERENCE = "speciesReference"
    STOICHIOMETRY = "stoichiometry"

    # listOfParameters
    PARAMETER = "localParameter"
    VALUE = "value"
    UNITS = "units"
    # Extensions for flux balance models
    P_LOWER_BOUND = "LOWER_BOUND"
    P_UPPER_BOUND = "UPPER_BOUND"
    P_OBJ_COEF = "OBJECTIVE_COEFFICIENT"
    P_FLUX_VALUE = "FLUX_VALUE"

    BIOMASS = "biomass"  # used by SbmlParser to recognize the biomass reaction


def getFirstElementChild(node):
    """ return the first XML child node that is of type Element
    """
    for child in node.childNodes:
        if child.nodeType == node.ELEMENT_NODE:
            return child



#!/usr/bin/env python

""" metano Metabolite Flux Minimization

This script performs MFM on the FBA solution for a metabolic network (given as
a set of reaction equations), an objective function, and a number of inequality
constraints.

It first performs a standard FBA (or reads in an FBA solution) and then performs
variability analysis by the following algorithm:

1. Introduce a new constraint: Original objective function must have a value
                               of at least 95% of the optimal value (as
                               determined from the original FBA; threshold is
                               configurable).
2. Generate coefficient vectors expressing the (producing) metabolite fluxes as
   linear combinations of the reaction fluxes (flux variables).
3. Perform LP with the linear objective functions defined by these coefficient
   vectors: Successively minimize each objective function.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke, Helge Stark
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function


from builtins import map
from builtins import range
from builtins import object
import os
import csv
from numpy import array, vstack, append, nan, isinf
from metano.defines import FbaParam, padNumber, COPYRIGHT_VERSION_STRING
from metano.fba import OptionParser, FbAnalyzer
from metano.reactionparser import ReactionParser
from metano.paramparser import ParamParser
from metano.metabolicmodel import MetabolicModel
from metano.linearproblem import LinearProblem

EPSILON = 1E-9
LARGE_NUM = 1E4


class MfmAnalyzer(object):
    """ Class for metabolite flux minimization
    """

    def __init__(self, solver="default"):
        """ initialize MFM class

        Keyword arguments:

        solver    -- LP solver to be used
        """
        self.solver = solver

    def run(self, objective, matrix, lb, ub, threshold=1.0, eqs=[], ineqs=[], minimize=True):
        """ successively formulate and solve linear problems for each metabolite
            flux with inequality constraint 'objective value <= threshold'

        Arguments refer to split fluxes (i.e. non-negative flux variables).

        Keyword arguments:

        objective      -- coefficient vector for linear objective function
        matrix         -- stoichiometric matrix of the metabolic network
        lb             -- list of lower bounds (indexed like matrix columns)
        ub             -- list of upper bounds (indexed like matrix columns)
        threshold      -- objective function threshold
        eqs            -- list of (coefficient vector, right-hand side) pairs
                          for additional equality constraints
        ineqs          -- list of - '' - for additional inequality constraints
        minimize       -- True for metabolite flux minimization, False for maximization (NEW)

        Returns:  list of minimum values, indexed like metabolites
        """
        matrix = array(matrix)
        nMetabolites, nCols = matrix.shape
        if nMetabolites == 0:
            return []  # Nothing to do

        # Construct matrices of original equality and inequality constraints
        Aeq, beq, Aineq, bineq = FbAnalyzer.makeConstraintMatrices(matrix, eqs,
                                                                   ineqs)
        # Combine original inequality constraints and
        # 'objective function <= threshold'
        if Aineq is None:
            Aineq = [objective]
            bineq = [threshold]
        else:
            Aineq = vstack((Aineq, [objective]))
            bineq = append(bineq, threshold)

        result = []
        ubVec = []
        for i in range(nCols):
            # Impose artificial upper bounds so that all LPs are bounded
            if isinf(ub[i]):
                ubVec.append(LARGE_NUM)
            else:
                ubVec.append(ub[i])

        psSplit = LinearProblem(Aeq, beq, Aineq, bineq, array(lb), array(ubVec),
                                self.solver)

        # Compute coefficient vectors of producing metabolite fluxes
        # - pick only positive coefficients from stoichiometric matrix
        posMatrix = matrix.copy()
        for i in range(nMetabolites):
            for j in range(nCols):
                if posMatrix[i, j] < 0.:
                    posMatrix[i, j] = 0.

        # Successively minimize each flux
        for index in range(nMetabolites):
            try:
                obj = list(map(float, posMatrix[index, :]))
                if not minimize:
                    obj = [-x for x in obj]
                psSplit.setObjective(obj)
            except Exception:
                # Skip all-zero rows
                result.append(0.)
                continue

            minval, s = psSplit.minimize()
            psSplit.resetObjective()

            if len(s) == 0:
                result.append(nan)
            else:
                result.append(minval)

        return result

    def runOnModel(self, model, fbaParams, threshp=.95, objFuncVal=None,
                   rmDeadEnds=True, minimize=True):
        """ perform metabolite flux minimization on the given model with
            objective value <= threshp * maximum

        Keyword arguments:

        model        -- the MetabolicModel
        fbaParams    -- FBA parameters
        threshp      -- threshold percentage (objective_value <= thresp*maximum)
        objFuncVal   -- FBA optimum for objective function (optional)
        rmDeadEnds   -- if True, remove all reactions with dead ends before
                        analysis (faster and gives an optimal solution, as well)

        Returns:
        minVec, dimReduced

        minVec       -- list of minimum values, indexed like metabolites
        dimReduced   -- pair (nRows, nColumns) with dimensions of reduced matrix
        """
        if rmDeadEnds:
            deadReactions = model.findDeadEnds(True)[1]
            modelRed = model.getSubModelByExcludeList(deadReactions)
            cbz = model.canBeZero(deadReactions)
            nonZeroDeadEnds = [deadReactions[i] for i in
                               range(len(deadReactions)) if not cbz[i]]
            if nonZeroDeadEnds:
                print ("The following blocked reactions are constrained to a "
                       "non-zero flux:\n  " + "\n  ".join(nonZeroDeadEnds) +
                       "\nThe problem is infeasible.")
                return [], array(modelRed.getStoichiometricMatrix()).shape
        else:
            modelRed = model

        matrix = array(modelRed.getStoichiometricMatrix())
        dimReduced = matrix.shape
        lb, ub = modelRed.getBounds()

        # Split fluxes into non-negative components
        matrixSplit, reactionsSplit, lbSplit, ubSplit = \
            FbAnalyzer.splitFluxes(matrix, modelRed.getReactionNames(), lb, ub)

        # Build (negative) objective function vector for split fluxes
        objective = ParamParser.convertObjFuncToLinVec(fbaParams.objStr,
                                                       reactionsSplit, len(lbSplit), fbaParams.maxmin)
        maxmin_factor = -1. if fbaParams.maxmin else 1.

        # If the optimum of the objective function is not given, perform FBA
        if objFuncVal is None:
            fba = FbAnalyzer(self.solver)

            objFuncVal, sFlux = fba.run(reactionsSplit, matrixSplit, lbSplit,
                                        ubSplit, fbaParams)

            if len(sFlux) == 0:
                return [], dimReduced

        # Use negative threshold (objective value >= threshold is equivalent to
        # -objective value <= -threshold, and objective already has coefficient
        # -1 due to maximization)
        threshold = maxmin_factor * objFuncVal * threshp
        print("obj func. opt:", objFuncVal)
        print("Threshold: ", threshold)
        try:
            eqs, ineqs = ParamParser.linConstraintsToVectors(
                fbaParams.linConstraints, modelRed.reactionDict, len(lbSplit))
        except ValueError as e:
            # If any linear constraint is contradictory, report error
            print("Optimization not possible due to contradictory constraints:")
            print("  "+e)
            exit()
        minVec = self.run(objective, matrixSplit, lbSplit, ubSplit, threshold,
                          eqs, ineqs, minimize)
        if not minimize:
            minVec = [-x for x in minVec]

        # Add removed reactions with min = 0. to minVec and solution
        if len(modelRed) != len(model):
            minVecFull = []
            metabolitesRed = set(modelRed.getMetaboliteNames())

            for met in model.getMetaboliteNames():
                if met in metabolitesRed:
                    # Take values from minVec
                    index = modelRed.metaboliteDict[met]
                    minVecFull.append(minVec[index])
                else:
                    # Dead end => min = 0.
                    minVecFull.append(0.)

            return minVecFull, dimReduced
        else:
            return minVec, dimReduced

    @staticmethod
    def writeSolutionToFile(filename, metabolites, minVec):
        with open(filename, 'w') as f:
            if filename.split(".")[-1] == "csv":
                MfmAnalyzer.writeCSVToFileHandle(f, metabolites, minVec)
            else:
                MfmAnalyzer.writeSolutionToFileHandle(f, metabolites, minVec)

    @staticmethod
    def writeCSVToFileHandle(f, metabolites, minVec):
        """ write MFM solution to the given CSV file handle (must be open for
            writing)"""
        if len(minVec) == 0:
            return  # Nothing to do
        if len(metabolites) != len(minVec):
            raise ValueError("Length mismatch: 'metabolites'(%u) 'minVec'(%u)"
                             % (len(metabolites), len(minVec)))

        csvwriter = csv.writer(
            f, delimiter=";", quotechar='"', quoting=csv.QUOTE_MINIMAL)
        # Write Header to CSV file
        csvwriter.writerow(["NAME", "MIN_FLUX"])
        # Construct output table and get widths of table columns (length of
        # longest occurring string)
        for met in sorted(metabolites):
            index = metabolites[met]
            csvwriter.writerow([met, minVec[index]])

    @staticmethod
    def writeSolutionToFileHandle(f, metabolites, minVec):
        """ write MFM solution to the given file handle (must be open for
            writing)

        Keyword arguments:

        metabolites  -- dictionary { name : index }
        minVec       -- list of minimum flux values for each metabolite
        """
        if len(minVec) == 0:
            return  # Nothing to do
        if len(metabolites) != len(minVec):
            raise ValueError("Length mismatch: 'metabolites'(%u) 'minVec'(%u)"
                             % (len(metabolites), len(minVec)))

        # Construct output table and get widths of table columns (length of
        # longest occurring string)
        maxlenName, maxlenVal = 0, 0
        output = {}  # dict {name : flux value as string }
        for met in metabolites:
            index = metabolites[met]
            # Pad non-negative numbers with space
            valStr = padNumber(repr(minVec[index]))
            output[met] = valStr
            lenName, lenVal = len(met), len(valStr)
            if lenName > maxlenName:
                maxlenName = lenName
            if lenVal > maxlenVal:
                maxlenVal = lenVal

        # Write table head
        f.write("NAME".ljust(maxlenName)+"   " +
                "MIN_FLUX".rjust(maxlenVal)+"\n")

        # Write result for the producing flux through every metabolite
        for met in sorted(metabolites):
            f.write(met.ljust(maxlenName)+" : "+output[met].rjust(maxlenVal) +
                    "\n")

    @staticmethod
    def parseSolutionFile(filename):
        """ read the given MFM solution file identified by filename
            -- wrapper for parseSolutionFileByHandle()
        """
        with open(filename) as f:
            if filename.split(".")[-1] == "csv":
                return MfmAnalyzer.parseCSVFileByHandle(f)
            else:
                return MfmAnalyzer.parseSolutionFileByHandle(f)

    @staticmethod
    def parseCSVFileByHandle(f):
        """ read the MFM solution file given as a file object (must be open)

        Returns:
        (fbaSolution, minmax) tuple with

        fbaSolution -- MetabolicFlux object with flux distribution, and lb, ub
        minmax      -- dict { reaction : flux minimum, flux maximum }
        """
        csvreader = csv.reader(f, delimiter=";", quotechar='"')
        reactions = set()
        reactionVec, fluxVec = [], []

        line_no = 0
        for line in csvreader:
            line_no += 1
            if line[0] == "NAME":
                continue

            rea = line[0]
            if rea in reactions:
                raise SyntaxError("Syntax error in line %u: Duplicate "
                                  "reaction." % line_no)
            reactions.add(rea)
            reactionVec.append(rea)

            try:
                values = list(map(float, line[1:]))
            except ValueError:
                raise SyntaxError("Syntax error in line %u: Invalid "
                                  "floating point value." % line_no)
            try:
                fluxVec.append(values[1])
            except IndexError:
                raise SyntaxError("Syntax error in line %u:\nLine must "
                                  "contain exactly one value (min_flux)" %
                                  line_no)
        return MetabolicFlux(reactionVec, fluxVec, boundsDict), minmax

    @staticmethod
    def parseSolutionFileByHandle(f):
        """ read the MFM solution file given as a file object (must be open)

        Returns:
        (fbaSolution, minmax) tuple with

        fbaSolution -- MetabolicFlux object with flux distribution, and lb, ub
        minmax      -- dict { reaction : flux minimum, flux maximum }
        """
        metabolites = set()
        mfmDict = {}

        line_no = 0
        for line in f:
            line_no += 1
            if (line.lstrip().upper().startswith("NAME") or line == "" or
                    line.isspace()):
                continue

            try:
                met, values_str = list(map(str.rstrip, line.split(":")))
            except ValueError:
                raise SyntaxError("Syntax error in line %u:\nLine must "
                                  "contain exactly one colon (':')." %
                                  line_no)
            if met in metabolites:
                raise SyntaxError("Syntax error in line %u: Duplicate "
                                  "reaction." % line_no)
            metabolites.add(met)

            try:
                value = float(values_str.strip())
            except ValueError:
                raise SyntaxError("Syntax error in line %u: Invalid "
                                  "floating point value." % line_no)
            mfmDict[met] = value

        return mfmDict


def main():
    # 1. Parse command line

    usage = "Usage: %prog [options]"
    version = "Metabolite flux minimization\n" + COPYRIGHT_VERSION_STRING
    parser = OptionParser(usage=usage, version=version)
    parser.add_option("-r", "--reactions", dest="reactionFile",
                      help="perform Metabolite Flux Minimization on the"
                      " network given by the reaction FILE", metavar="FILE")
    parser.add_option("-p", "--parameters", dest="paramFile",
                      help="use the given scenario FILE", metavar="FILE")
    parser.add_option("-o", "--output", dest="outputFile",
                      help="write MFM output to FILE", metavar="FILE")
    parser.add_option("-t", "--tolerance", dest="tolerance", type="float",
                      help="tolerance for objective function (0 < VALUE "
                      "< 1; default: .95)", metavar="VALUE")
    parser.add_option("-l", "--use-full-matrix", action="store_true",
                      dest="useFullMatrix", help="use full matrix (disable "
                      "removal of dead ends and nonfunctional reactions)")
    parser.add_option("-m", "--maximize", action="store_true",
                      dest="maximize", help="maximize metabolite flux "
                      "instead of minimization")
    parser.set_defaults(tolerance=.95, useFullMatrix=False, maximize=False)

    options, _ = parser.parse_args()
    parser.check_required("-r")
    parser.check_required("-p")
    parser.check_required("-o")

    if options.tolerance <= 0. or options.tolerance > 1.:
        print("Error: Tolerance must be in interval (0, 1]")
        exit()

    # 2. Parse reaction file

    rparser = ReactionParser()
    model = MetabolicModel()
    try:
        model.addReactionsFromFile(options.reactionFile, rparser)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in reaction file %s:" %
               os.path.basename(options.reactionFile))
        print(strerror)
        exit()

    # 3. Parse scenario file

    model_messages = []
    pparser = ParamParser()
    try:
        # Parse file, get maxmin, name of objective function, and solver name
        maxmin, objStr, solver, numIter, lb, ub = pparser.parse(
            options.paramFile)
        fbaParams = FbaParam(solver, maxmin, objStr, numIter)
        fbaParams.setLinConstraints(pparser.lin_constraints)
        # Set flux bounds in model
        model.setFiniteBounds(lb, ub, True, model_messages)
    except IOError as strerror:
        print ("An error occurred while trying to read file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except SyntaxError as strerror:
        print ("Error in scenario file %s:" %
               os.path.basename(options.paramFile))
        print(strerror)
        exit()
    except ValueError as strerror:
        print(strerror)
        exit()

    # Show warning and info messages of parsers
    msgs = (rparser.getMessages() + pparser.getMessages() +
            [x[1] for x in model_messages])
    if msgs:
        print('\n'+'\n'.join(msgs))

    # 4. Launch metabolite flux minimization

    print("Launching metabolite flux minimization.")
    print("tolerance: %g" % options.tolerance)
    if solver == "":
        solver = "default"
    mfm = MfmAnalyzer(solver)
    minVec, ndim = mfm.runOnModel(model, fbaParams, options.tolerance,
                                  rmDeadEnds=not options.useFullMatrix,
                                  minimize=not options.maximize)

    print ("Info: The reduced network has %u reactions and %u metabolites." %
           ndim[1::-1])
    if not minVec:
        print("Model is infeasible or unbounded. Nothing to do.")
        exit()

    # 5. Write output to file

    try:
        mfm.writeSolutionToFile(options.outputFile, model.metaboliteDict,
                                minVec)
    except IOError as strerror:
        print ("Unable to write to file %s:" %
               os.path.basename(options.outputFile))
        print(strerror)
        exit()


if __name__ == "__main__":
    main()



"""
This module defines the class MetabolicModel, which stores a stoichiometric
metabolic model.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

Rene Rex contributed to this file.

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import print_function
from __future__ import absolute_import

from future import standard_library
standard_library.install_aliases()
from builtins import map
from builtins import zip
from builtins import range
from builtins import object
from metano.defines import Verbosity, ReaFileStruc, typename, makeUnique
from numpy import inf, array, nonzero
from io import StringIO
from metano.reactionparser import ReactionParser
import re
import codecs


class ModelError(Exception):
    """ Exception raised if an inconsistency in the model is detected
    """
    def __init__(self, value):
        self.value = value
    def __str__(self):
        return repr(self.value)


class Reactant(object):
    """ Class for a metabolite occurring in a reaction, stored with coefficient

        Member variables:

        name -- the metabolite's name
        coef -- the stoichiometric coefficient of the metabolite in the reaction
                (< 0: consumed, > 0: produced)
    """
    def __init__(self, name="", coef=0.):
        self.name = name
        self.coef = coef

    @staticmethod
    def reactantFromIterable(x):
        """ return a reactant from iterable x: coef <-- x[0], name <-- x[1];
            if x is already a Reactant, return x
        """
        if isinstance(x, Reactant):
            return x
        try:
            return Reactant(*x[1::-1])

        except TypeError:
            raise TypeError("Expected iterable or Reactant. Got '%s'" %
                            typename(x))
        except ValueError:
            raise TypeError("Expected iterable with at least two "
                            "elements")

    def __repr__(self):
        return "Reactant(%r, %r)" % (self.name, self.coef)

    def __iter__(self):
        """ return an iterator over the tuple (coef, name) - for conversion to
            tuple
        """
        return iter((self.coef, self.name))


class Reaction(object):
    """ Represents a single chemical reaction

        Member variables:

        name        -- the reaction's identifier
        direction   -- 0: reversible, 1: left to right, -1: right to left
        metabolites -- list of Reactant objects with name and coefficient
                       (must never be changed directly!); no two elements have
                       the same name (coef < 0: left-hand side,
                                      coef > 0: right-hand side)
        lb          -- lower bound for flux (-inf means unbounded) (property)
        ub          -- upper bound for flux (inf means unbounded)  (property)
        comment     -- comment string, single- or multi-line

        _metabDict  -- dictionary {name : index in metabolites}
        _lb, _ub    -- variables for lower and upper flux bound
    """

    # Mapping from arrow to corresponding direction enum value
    arrowToDirection = {ReaFileStruc.arrowRev: 0, ReaFileStruc.arrowIrr: 1,
                        ReaFileStruc.arrowFlip: -1}
    directionToArrow = dict(list(zip(list(arrowToDirection.values()),
                                list(arrowToDirection.keys()))))

    def __init__(self, name="", educts=None, arrow=None, products=None,
                 comment="", educt_factor=-1., lb=-inf, ub=inf, **kwargs):
        """ create a new reaction

        Keyword arguments:

        name      -- the reaction's identifier
        educts    -- reactants on left-hand side (list of Reactant objects or
                     list of (coef, name) pairs)
        arrow     -- <=>/-->/<--, specifies direction and reversibility
        products  -- reactants on right-hand side (list of Reactant objects or
                     list of (coef, name) pairs))
        educt_factor
                  -- factor for educts coefficients (default -1. assumes that
                     educts have positive signs)
        comment   -- comment string
        lb        -- lower bound for flux (-inf means unbounded)
        ub        -- upper bound for flux (inf means unbounded)

        direction -- reaction direction (alternatively to arrow)
        """
        self.name, self.comment, self._lb, self._ub = name, comment, lb, ub

        if not arrow:
            self.setDirection(kwargs.get("direction", 0))
            # If no direction or arrow is given, reaction is assumed reversible
        else:
            try:
                self.setDirection(self.arrowToDirection[arrow])
                # Arrow takes precedence over a possible "direction" keyword
                # argument
            except KeyError:
                raise ValueError("Illegal reaction arrow '%s'. Must be one of "
                                 "%s" % (arrow, list(self.arrowToDirection.keys())))

        self.metabolites = []
        self._metabDict = {}
        # Convert educts and products to Reactants if given as iterables
        for e in map(Reactant.reactantFromIterable, educts):
            self.addReactant(e, educt_factor)
        for p in map(Reactant.reactantFromIterable, products):
            self.addReactant(p)


    def clear(self):
        """ clear the Reaction object
        """
        self.name = ""
        self.direction = 0
        self.metabolites = []
        self._metabDict = {}
        self.comment = ""
        self._lb = -inf
        self._ub =  inf

    def addReactant(self, reactant, factor=1.):
        """ add the given reactant to the reaction, with its coefficient
            multiplied by the given factor
        """
        # If metabolite is already present, combine reactants (add coefficients)
        if reactant.name in self._metabDict:
            self.metabolites[self._metabDict[reactant.name]].coef += \
                reactant.coef*factor
        else:
            new_index = len(self.metabolites)
            self._metabDict[reactant.name] = new_index
            self.metabolites.append(reactant)
            self.metabolites[new_index].coef *= factor

    def removeReactant(self, name):
        """ remove the reactant given by name from the reaction
        """
        if name not in self._metabDict:
            raise KeyError(name+" not found in reaction "+self.name)

        del self.metabolites[self._metabDict[name]]
        # Update indices in dictionary
        for i in range(self._metabDict[name], len(self.metabolites)):
            self._metabDict[self.metabolites[i].name] = i
        del self._metabDict[name]


    def removeZeroReactants(self, tolerance=0.):
        """ remove all reactants with coefficients that are less than the given
            tolerance from zero
        """
        i = len(self.metabolites)-1
        while i >= 0:
            coef, name = self.metabolites[i]
            if abs(coef) <= tolerance:
                self.removeReactant(name)
            i -= 1


    def setDirection(self, direction):
        """ set reaction direction (either as arrow or as direction enum)
        """
        if direction in self.arrowToDirection:
            self.direction = self.arrowToDirection[direction]
        elif direction in self.directionToArrow:
            self.direction = direction
        else:
            raise ValueError("Illegal direction value or arrow: %r" % direction)

        # Adjust bounds if necessary
        if self.direction == 1:
            self._lb = max(self._lb, 0.)
            self._ub = max(self._ub, 0.)
        elif self.direction == -1:
            self._lb = min(self._lb, 0.)
            self._ub = min(self._ub, 0.)


    def getEducts(self, withZero=False):
        """ return the educts (i.e. reactants with coefficient < 0) as Reactant
            objects; if withZero is True, include reactants with coefficient 0
        """
        if withZero:
            return [x for x in self.metabolites if x.coef <= 0.]
        else:
            return [x for x in self.metabolites if x.coef < 0.]

    def getProducts(self, withZero=False):
        """ return the products (i.e. reactants with coefficient > 0);
            if withZero is True, include reactants with coefficient 0
        """
        if withZero:
            return [x for x in self.metabolites if x.coef >= 0.]
        else:
            return [x for x in self.metabolites if x.coef > 0.]

    def setBounds(self, lb, ub):
        self.lb = lb
        self.ub = ub

    def getBounds(self):
        return self._lb, self._ub

    def getLb(self):
        """ lower bound for the flux through the reaction
        """
        return self._lb

    def setLb(self, val):
        """ set lower bound for flux, adjusted for reaction direction
        """
        if self.direction == 0:
            self._lb = val
        elif self.direction == 1:
            self._lb = max(val, 0.)
        else:
            self._lb = min(val, 0.)

    def getUb(self):
        """ upper bound for the flux through the reaction
        """
        return self._ub

    def setUb(self, val):
        """ set lower bound for flux, adjusted for reaction direction
        """
        if self.direction == 0:
            self._ub = val
        elif self.direction == 1:
            self._ub = max(val, 0.)
        else:
            self._ub = min(val, 0.)

    lb = property(getLb, setLb)
    ub = property(getUb, setUb)


    def writeToFileHandle(self, f, reactionLabels=None, metaboliteLabels=None,
                          width=0, addComment=False, addNewLine=True):
        """ write the reaction equation to the given file handle, replace
            reaction name and reactant names with labels if given

            Keyword arguments:

            f                -- file object (must be open for writing)
            reactionLabels   -- either a single label or dict {name : label}
            metaboliteLabels -- dict {name : label} for metabolites
            width            -- string width for alignment of reaction name
            addComment       -- if True, append '  # <comment string>' to output
            addNewLine       -- if True, end output with newline character
        """
        if reactionLabels:
            try:
                reaname = reactionLabels[self.name]
            except TypeError:
                reaname = reactionLabels
        else:
            reaname = self.name

        s = ""
        if addComment and self.comment and self.comment.find('\n') >= 0:
            multiLine = True
            # Display multi-line comment before reaction
            for line in self.comment.split('\n'):
                s += ReaFileStruc.commentSign + " " + line + '\n'
        else:
            multiLine = False

        # Only show reactants with non-zero coefficients
        educts, products = [], []
        if metaboliteLabels:
            for coef, name in self.metabolites:
                if coef < 0.:
                    educts.append("%g %s" % (-coef,
                                  metaboliteLabels.get(name, name)))
                elif coef > 0.:
                    products.append("%g %s" % (coef,
                                    metaboliteLabels.get(name, name)))
        else:
            for coef, name in self.metabolites:
                if coef < 0.:
                    educts.append("%g %s" % (-coef, name))
                elif coef > 0.:
                    products.append("%g %s" % (coef, name))

        arrow = self.directionToArrow[self.direction]
        s += "%s %s %s %s %s" % (reaname.ljust(width), ReaFileStruc.delimiter,
                                " + ".join(educts), arrow, " + ".join(products))
        if addComment and not multiLine and self.comment:
            s += "  %s %s" % (ReaFileStruc.commentSign, self.comment)
        if addNewLine:
            s += '\n'
        f.write(s)

    def toString(self, includeZeroReactants=False):
        """ generate a string representation of the reaction for output

        If includeZeroReactants is False, reactants with coefficient zero are
        omitted.
        """
        educts, products = [], []
        for coef, name in self.metabolites:
            # Only display reactants with non-zero coefficients
            if coef < 0.:
                educts.append("%g %s" % (-coef, name))
            elif coef > 0.:
                products.append("%g %s" % (coef, name))
            elif includeZeroReactants:
                educts.append("%g %s" % (0., name))
        arrow = self.directionToArrow[self.direction]
        return (self.name + " "+ReaFileStruc.delimiter+" " + " + ".join(educts)
                + " "+arrow+" " + " + ".join(products))


    def __str__(self):
        return self.toString()

    def __repr__(self):
        educts = [(-coef, name) for (coef, name) in self.getEducts(True)]
        products = list(map(tuple, self.getProducts(False)))
        return ("Reaction(%r, %r, %r, %r, %r, %r, %r, %r)" %
                (self.name, educts, self.directionToArrow[self.direction],
                 products, self.comment, -1., self._lb, self._ub))

    def __iter__(self):
        return iter(self.metabolites)

    def __len__(self):
        return len(self.metabolites)


class MetabolicModel(object):
    """ Represents a metabolic model as list of reactions and keeps track of all
        metabolites in the model.

    Member variables:

    reactions       -- a list of Reaction objects
    reactionDict    -- dictionary {name : index in reactions list}
    metabolites     -- a list of metabolite names
    metaboliteDict  -- dictionary {name : index in metabolites list}

    biomass_name        -- name of biomass reaction (None - not set)
    cofactors           -- list of cofactor metabolites
    _reactionLabels     -- dict {reaction name : label} for displaying the model
    _metaboliteLabels   -- dict {metabolite name : label} for displaying the
                           model
    useReactionLabels   -- if True, use _reactionLabels instead of reaction
                           names when displaying the model
    useMetaboliteLabels -- if True, use _metaboliteLabels instead of metabolite
                           names when displaying the model
    showComments        -- if True, display comments when displaying the model
                           (False by default)
    """

    def __init__(self, cofactors=set(), reactionLabels={}, metaboliteLabels={}):
        """ create a new MetabolicModel

        Keyword arguments:

        cofactors        -- a set of metabolites marked as cofactors
        reactionLabels   -- a dictionary of labels to be used in place of the
                            reaction names when displaying the model
        metaboliteLabels -- a dictionary of labels to be used in place of the
                            metabolite names when displaying the model

        Use of labels can be triggered at any time via the Boolean flags
        self.useReactionLabels and self.useMetaboliteLabels.
        """
        self.clearReactions()
        self.cofactors = cofactors
        self.setReactionLabels(reactionLabels)
        self.setMetaboliteLabels(metaboliteLabels)
        self.showComments = False


    def setReactionLabels(self, reactionLabels):
        """ set the reaction labels, which are used in place of the reaction
            names when displaying the model (given as dictionary {name : label})
        """
        self._reactionLabels = reactionLabels
        self.useReactionLabels = bool(reactionLabels)

    def setMetaboliteLabels(self, metaboliteLabels):
        """ set the metabolite labels, which are used in place of the metabolite
            names when displaying the model (given as dictionary {name : label})
        """
        self._metaboliteLabels = metaboliteLabels
        self.useMetaboliteLabels = bool(metaboliteLabels)

    def makeLabelsUnique(self, doReactionLabels=True, doMetaboliteLabels=True):
        """ generate unique labels for reactions, metabolites, or both
            - identical labels are distinguished by appending numbers
        """
        if doReactionLabels:
            reanames = self.getReactionNames()
            if len(self.reactions) != len(set(reanames)):
                raise ModelError("Reaction names are not unique")
            self._reactionLabels = dict(list(zip(reanames,
                makeUnique(self._getTranslatedReactionNames()))))

        if doMetaboliteLabels:
            # Metabolite names are unique by construction

            self._metaboliteLabels = dict(list(zip(self.getMetaboliteNames(),
                makeUnique(self._getTranslatedMetaboliteNames()))))


    def getBounds(self):
        """ return the bounds of all reaction fluxes as a pair of lists, indexed
            like the list of reactions (and thus like the columns of the
            stoichiometric matrix); -inf/inf means that a reaction is unbounded
            to the left/right side
        """
        lbVec, ubVec = [], []
        for r in self.reactions:
            lbVec.append(r.lb)
            ubVec.append(r.ub)
        return lbVec, ubVec

    def setBounds(self, lbVec=None, ubVec=None):
        """ set lower and upper bounds of all reaction fluxes - both lists are
            indexed like the list of reactions (and thus like the columns of the
            stoichiometric matrix); -inf/inf means that a reaction is unbounded
            to the left/right side

        Keyword arguments:

        lbVec -- vector of lower bounds (may be empty)
        ubVec -- vector of upper bounds (may be empty)
        """
        nReactions = len(self.reactions)
        if lbVec is not None:
            if len(lbVec) != nReactions:
                raise ValueError("lbVec has wrong length (%u); expected %u" %
                                 (len(lbVec), nReactions))
            for i in range(nReactions):
                self.reactions[i].lb = lbVec[i]
        if ubVec is not None:
            if len(ubVec) != nReactions:
                raise ValueError("ubVec has wrong length (%u); expected %u" %
                                 (len(ubVec), nReactions))
            for i in range(nReactions):
                self.reactions[i].ub = ubVec[i]

    def getFiniteBounds(self):
        """ return the finite bounds of all reaction fluxes as a pair of dicts
            {reaction name : value}
        """
        lbDict, ubDict = {}, {}
        for r in self.reactions:
            lb, ub = r.lb, r.ub
            if lb > -inf:
                lbDict[r.name] = lb
            if ub < inf:
                ubDict[r.name] = ub
        return lbDict, ubDict

    def setFiniteBounds(self, lbDict={}, ubDict={}, strict=True, messages=[]):
        """ set lower and upper bounds of fluxes through specific reactions

        Keyword arguments:

        lbDict   -- dict {reaction name : value} for lower bounds
        ubDict   -- dict {reaction name : value} for upper bounds
        strict   -- if True, raise ValueError if both bounds are outside the
                    range allowed by the reaction's direction
        messages -- list in which to store warnings and info messages (as
                    (level, msg) pairs)
        """
        for reaname in ubDict:
            try:
                index = self.reactionDict[reaname]
            except KeyError:
                msg = "Warning: UB set for non-existing reaction %s" % reaname
                messages.append((Verbosity.WARNING, msg))
                continue

            if self.reactions[index].direction == 1 and ubDict[reaname] < 0.:
                msg = ("Error: Upper bound for irreversible reaction %s is "
                       "negative" % reaname)
                if strict:
                    raise ValueError(msg)
                else:
                    messages.append((Verbosity.ERROR, msg))
            self.reactions[index].ub = ubDict[reaname]

        for reaname in lbDict:
            try:
                index = self.reactionDict[reaname]
            except KeyError:
                msg = "Warning: LB set for non-existing reaction %s" % reaname
                messages.append((Verbosity.WARNING, msg))
                continue

            if self.reactions[index].direction == -1 and lbDict[reaname] > 0.:
                msg = ("Error: Lower bound for right-to-left irreversible "
                       "reaction %s is negative" % reaname)
                if strict:
                    raise ValueError(msg)
                else:
                    messages.append((Verbosity.ERROR, msg))
            self.reactions[index].lb = lbDict[reaname]


    def getStoichiometricMatrix(self):
        """ compute the stoichiometric matrix from the list of reactions

        Rows are indexed like the metabolite list, columns are indexed like the
        reaction list.
        """
        nReactions = len(self.reactions)
        matrix = [[0.]*nReactions for _ in self.metaboliteDict]
        for j in range(nReactions):
            for coef, name in self.reactions[j]:
                matrix[self.metaboliteDict[name]][j] = coef

        return matrix


    def clearReactions(self):
        """ clear the network, i.e. remove all reactions and metabolites
        """
        self.reactions = []
        self.reactionDict = {}
        self.metabolites = []
        self.metaboliteDict = {}
        self.biomass_name = None


    def addReactions(self, reactions):
        """ append the given list of reactions to the model

            Reactions are given in one of the following formats:
            a) Reaction objects
            b) tuples (name, educts, arrow, products[, comment]) with educts and
               products either as Reactant objects or (coef, name) tuples
            c) tuples (name, educts, arrow, products, comment, lb, ub)
            In cases b) and c), educts must have positive coefficients.
        """
        first_new = len(self.reactions)
        reactionSet = set(self.reactionDict.keys())
        for r in reactions:
            if isinstance(r, Reaction):
                if r.name in reactionSet:
                    raise ModelError("Reaction '%s' already in model" % r.name)
                reactionSet.add(r.name)
                self.reactions.append(r)
            else:
                try:
                    numElements = len(r)
                except TypeError:
                    raise TypeError("Expected tuple or Reaction, got '%s'" %
                                    typename(r))
                if numElements < 4:
                    raise TypeError("Reaction tuple has wrong length - must be "
                                    "at least (name, educts, arrow, products)")
                reaname = r[0]
                if reaname in reactionSet:
                    raise ModelError("Reaction '%s' already in model" % reaname)
                reactionSet.add(reaname)
                if numElements < 6:
                    self.reactions.append(Reaction(*r[:5]))
                else:
                    self.reactions.append(Reaction(*(r[:5]+(-1.,)+r[5:7])))

            # If any reaction has name 'biomass' (case-independent), mark it as
            # biomass reaction if biomass_name is not set otherwise
            if (self.reactions[-1].name.lower() == "biomass" and
                not self.biomass_name):
                self.biomass_name = self.reactions[-1].name

        # Update reactionDict, metabolites, and metaboliteDict
        nMetabolites = len(self.metabolites)
        for i in range(first_new, len(self.reactions)):
            reaction = self.reactions[i]
            self.reactionDict[reaction.name] = i

            for reactant in reaction:
                met = reactant.name
                if met not in self.metaboliteDict:
                    self.metaboliteDict[met] = nMetabolites
                    self.metabolites.append(met)
                    nMetabolites += 1


    def addReaction(self, reaction):
        """ add a single reaction to the model using addReactions - see
            definition of addReactions for a description of the input format
        """
        self.addReactions([reaction])


    def addReactionsFromFile(self, filename, rparser=None):
        """ parse the file given by filename; if no ReactionParser is supplied,
            one is created for the occasion
        """
        if not rparser:
            rparser = ReactionParser()
        self.addReactions(rparser.parse(filename))


    def addReactionsFromFileHandle(self, f, rparser=None):
        """ parse the file given by file handle f
        """
        if not rparser:
            rparser = ReactionParser()
        self.addReactions(rparser.parseByHandle(f))

    def addReactionsFromString(self, s, rparser=None):
        """ parse the reaction file contents given in string s
        """
        if not rparser:
            rparser = ReactionParser()
        reafile = StringIO(s+'\n')
        try:
            reactions = rparser.parseByHandle(reafile)
        finally:
            reafile.close()
        self.addReactions(reactions)

    def addReactionsFromList(self, reactionList, rparser=None):
        """ parse the given list of reaction strings
        """
        self.addReactionsFromString('\n'.join(reactionList), rparser)


    def removeReaction(self, r, doUpdate=True):
        """ remove the reaction given by name, by index, or as Reaction object
            from the list; metabolites and metaboliteDict are only updated if
            doUpdate is True
        """
        if isinstance(r, Reaction):
            name = r.name
            if name not in self.reactionDict:
                raise KeyError(name+" not found in model")
            index = self.reactionDict[name]
        elif isinstance(r, str):
            name = r
            if name not in self.reactionDict:
                raise KeyError(name+" not found in model")
            index = self.reactionDict[name]
        else:
            index = r
            try:
                name = self.reactions[index].name
            except TypeError:
                raise TypeError("Expected Reaction, str, or int, got '%s'" %
                                typename(r))

        del self.reactions[index]
        del self.reactionDict[name]
        # Update reactionDict
        for i in range(index, len(self.reactions)):
            self.reactionDict[self.reactions[i].name] = i

        if doUpdate:
            self._updateMetaboliteList()


    def _updateMetaboliteList(self):
        """ build metabolites list and metaboliteDict from reactions list

        This function is called internally by removeReaction(). It preserves the
        order of the metabolites list. However, a call to removeReaction()
        followed by addReactions() will most likely change the order.
        Note: This function only accesses the reactions list, not reactionDict.
        """
        # Set of all metabolites in all reactions (for checking)
        metabSet = set()
        # List of all metabolites in all reactions (for order)
        metabolites = []
        for reaction in self.reactions:
            for _, met in reaction:  # get only names, ignore coefficients
                if met not in metabSet:
                    metabSet.add(met)
                    metabolites.append(met)

        # Remove all metabolites from model that are not present in metabSet
        nMetabolites = len(self.metabolites)
        first_deleted = nMetabolites
        for i in range(nMetabolites-1, -1, -1):
            met = self.metabolites[i]
            if met not in metabSet:
                del self.metabolites[i]
                del self.metaboliteDict[met]
                first_deleted = i

        # Update self.metaboliteDict
        nMetabolites = len(self.metabolites)
        for i in range(first_deleted, nMetabolites):
            self.metaboliteDict[self.metabolites[i]] = i
        # As this function is called only after removing reactions, we don't
        # need to check for metabolites in list 'metabolites' that are not
        # present in self.metabolites.


    def removeReactionsByNameList(self, nameList, doUpdate=True):
        """ remove all reactions specified by name in nameList;
            metabolites and metaboliteDict are only updated if doUpdate is True

        Names that don't correspond to any reaction in the model are ignored.
        """
        for name in nameList:
            if name in self.reactionDict:
                self.removeReaction(name, False)
        if doUpdate:
            self._updateMetaboliteList()

    def removeReactionsByList(self, reactionList, doUpdate=True):
        """ remove all reactions specified by the strings (reaction equations)
            in reactionList;
            metabolites and metaboliteDict are only updated if doUpdate is True

        Reactions are identified by name only, equations are not checked.
        """
        rparser = ReactionParser()
        reafile = StringIO('\n'.join(reactionList)+'\n')
        try:
            for reaction_entry in rparser.parseByHandle(reafile):
                self.removeReaction(reaction_entry[0], False)
        finally:
            reafile.close()
        if doUpdate:
            self._updateMetaboliteList()

    def removeReactionsByModel(self, model, doUpdate=True):
        """ remove all reactions present in both model and self from self;
            metabolites and metaboliteDict are only updated if doUpdate is True

        Reactions are identified by name only, Reactants are not checked.
        """
        for reaction in model:
            if reaction.name in self.reactionDict:
                self.removeReaction(reaction, False)
        if doUpdate:
            self._updateMetaboliteList()


    def getSubmodelByList(self, nameList):
        """ construct a submodel from the reactions identified by nameList

        Names that don't correspond to any reaction in the model are ignored.
        """
        m = MetabolicModel(self.cofactors, self._reactionLabels,
                           self._metaboliteLabels)
        # Collect all reactions whose names are in nameList
        rlist = []
        for name in nameList:
            if name in self.reactionDict:
                rlist.append(self.reactions[self.reactionDict[name]])
        m.addReactions(rlist)
        return m


    def getSubModelByExcludeList(self, nameList):
        """ construct a submodel from all reactions in the model not in nameList

        Names that don't correspond to any reaction in the model are ignored.
        """
        m = MetabolicModel(self.cofactors, self._reactionLabels,
                           self._metaboliteLabels)
        # Collect all reactions whose names are not in nameList
        nameSet = set(nameList)
        rlist = []
        for rea in self.reactions:
            if rea.name not in nameSet:
                rlist.append(rea)
        m.addReactions(rlist)
        return m


    def getSubmodelByRegex(self, pattern):
        """ construct submodel from the reactions identified by the name pattern

        Pattern can be either a single regular expression (given as string) or a
        list of regular expressions.
        """
        if isinstance(pattern, str):
            pattern = [pattern]

        regex = "|".join("^.*"+p+".*$" for p in pattern)
        matchedReactions = re.findall(re.compile(regex, re.MULTILINE),
                                        self.writeToString(False))

        m = MetabolicModel(self.cofactors, self._reactionLabels,
                           self._metaboliteLabels)
        m.addReactionsFromList(matchedReactions)
        # Add bounds and comments lost due to exporting and reading reactions
        # from string
        if self.useReactionLabels:
            # Create reverse mapping of labels to reaction names
            revMap = dict(list(zip(list(self._reactionLabels.values()),
                              list(self._reactionLabels.keys()))))
        else:
            rNames = self.getReactionNames()
            revMap = dict(list(zip(rNames, rNames)))
        for rea in m.reactions:
            original = self.reactions[self.reactionDict[revMap[rea.name]]]
            rea.setBounds(*original.getBounds())
            rea.comment = original.comment

        return m


    def getReactionNames(self):
        """ return the names (not labels) of the reactions in the model (in
            correct order)
        """
        return [r.name for r in self.reactions]

    def getMetaboliteNames(self):
        """ return the names (not labels) of the metabolites in the model (in
            correct order)
        """
        return self.metabolites


    def compare(self, otherModel):
        diffModel = MetabolicModel()
        # Compute difference based on line-by-line comparison of exported models
        self_export = self.writeToString(False).split('\n')
        other_export = otherModel.writeToString(False).split('\n')
        diffModel.addReactionsFromList(list(
            set(self_export).difference(other_export)))
        return diffModel


    def _getTranslatedReactionNames(self):
        if self.useReactionLabels:
            return [self._reactionLabels.get(reaction.name,
                                             reaction.name).replace(" ", "_")
                    for reaction in self.reactions]
        else:
            return self.getReactionNames()

    def _getTranslatedMetaboliteNames(self):
        if self.useMetaboliteLabels:
            return [self._metaboliteLabels.get(met, met).replace(" ", "_")
                    for met in self.metabolites]
        else:
            return self.getMetaboliteNames()

    def writeToFileHandle(self, outputFile, withComments=True,
                          useReactionLabels=None, useMetaboliteLabels=None,
                          sort=False):
        """ write the model (as reaction file) to the given file handle

        Keyword arguments:

        outputFile          -- file object (must be open for writing)
        withComments        -- if True, include reaction comments
        useReactionLabels   -- if not None, self.~ is set to the given value
        useMetaboliteLabels -- if not None, self.~ is set to the given value
        sort                -- if True, the reactions will be sorted by reaction
                               name
        """
        if not self.reactions:
            return  # Nothing to do

        if useReactionLabels is not None:
            self.useReactionLabels = useReactionLabels
        if useMetaboliteLabels is not None:
            self.useMetaboliteLabels = useMetaboliteLabels

        reactionLabels = self._getTranslatedReactionNames()
        maxlen = len(max(reactionLabels, key=len))
        rlabels = dict(list(zip(self.getReactionNames(), reactionLabels)))
        mlabels = dict(list(zip(self.getMetaboliteNames(),
                           self._getTranslatedMetaboliteNames())))

        last_index = len(self.reactions)-1
        count = 0

        if sort:

            for reactionName in sorted(self.getReactionNames()):
                reaction = self.reactions[self.reactionDict[reactionName]]
                reaction.writeToFileHandle(outputFile, rlabels, mlabels, maxlen,
                                           withComments, count < last_index)
                count += 1
        else:

            for reaction in self.reactions:
                reaction.writeToFileHandle(outputFile, rlabels, mlabels, maxlen,
                                           withComments, count < last_index)
                count += 1

    def writeToFile(self, outputFilename, withComments=True,
                    useReactionLabels=None, useMetaboliteLabels=None,
                    sort=False):
        """ write the model (as reaction file) to the given file using
            self.writeToFileHandle()
        """
        with codecs.open(outputFilename, 'w', encoding='utf-8') as outputFile:
            self.writeToFileHandle(outputFile, withComments, useReactionLabels,
                                   useMetaboliteLabels, sort)
            outputFile.write('\n')

    def writeToString(self, withComments=None, useReactionLabels=None,
                      useMetaboliteLabels=None):
        """ return a string representation of the model (as reaction file);

        Keyword arguments:

        withComments        -- if True, include reaction comments;
                               if None, use self.showComments for decision
        useReactionLabels   -- if not None, self.~ is set to the given value
        useMetaboliteLabels -- if not None, self.~ is set to the given value
        """
        outputFile = StringIO()
        if withComments is None:
            with_comments = self.showComments
        else:
            with_comments = withComments

        self.writeToFileHandle(outputFile, with_comments, useReactionLabels,
                               useMetaboliteLabels)

        output = outputFile.getvalue()
        outputFile.close()
        return output

    def __str__(self):
        return self.writeToString()


    def __iter__(self):
        return iter(self.reactions)

    def __len__(self):
        return len(self.reactions)


    def writeBoundsToParamFileHandle(self, f, useReactionLabels=None):
        """ write all bounds as "LB|UB <reaction name> <value>" lines to the
            given file handle (expected to be open for writing)

        Keyword arguments:

        f                   -- file object (must be open for writing)
        useReactionLabels   -- if not None, self.~ is set to the given value
        """
        if not self.reactions:
            return  # Nothing to do

        if useReactionLabels is not None:
            self.useReactionLabels = useReactionLabels

        reactionLabels = self._getTranslatedReactionNames()
        rlabels = dict(list(zip(self.getReactionNames(), reactionLabels)))

        for rea in self.reactions:
            lb, ub = rea.getBounds()
            if lb > -inf:
                f.write("LB %s %r\n" % (rlabels[rea.name], lb))
            if ub < inf:
                f.write("UB %s %r\n" % (rlabels[rea.name], ub))

    # ------ Exporters for different formats -----------------------------------

    def _formatReactantAnnetCsv(self, coef, name, is_ext, epsilon=1E-8):
        """ format a single reactant given with coefficient for output in anNET
            format (called internally by toAnnetCsvByHandle())

        Keyword arguments:
        coef    -- coefficient
        name    -- reactant's name
        is_ext  -- if True, mark as external, if False mark as cytosolic,
                   otherwise don't mark
        epsilon -- minimum float to be considered different from zero
        """
        # If coefficient is 1, omit it
        if abs(coef-1.) >= epsilon:
            s = "(%r) %s" % (coef, name)
        else:
            s = name
        try:
            c = {True : 'e', False : 'c'}[is_ext]
            s += "[%c]" % c
        except KeyError:
            pass
        return s

    def toAnnetCsv(self, filename, epsilon=1E-8):
        """ export the model as CSV file in anNET format to the file given by
            filename
        """
        with open(filename, 'w') as f:
            self.toAnnetCsvByHandle(f, epsilon)

    def toAnnetCsvByHandle(self, f, epsilon=1E-8):
        """ export the model as CSV file in anNET format to the given file

        Keyword arguments:
        f         -- file object (must be open for writing)
        epsilon   -- threshold below which a float is to be considered zero;
                     must be positive!
        """
        nMetabolites = len(self.metabolites)
        if nMetabolites == 0:
            return      # nothing to do

        # Replace parentheses with underscores in metabolite names
        metabReplace = dict(list(zip(self.metabolites,
                                [met.replace('(', '_').replace(')', '_')
                                 for met in self.metabolites])))

        # Write E. coli compartment data (must be adapted manually)
        f.write(";ID;pH;IS;Potential mV;Volume;\n"
                "compartment;e;3.5;0.15;0;0;extracellular\n"
                "compartment;c;6.5;0.15;0;1;cytosol\n\n;Model;;;;;\n\n"
                ";Abbreviation;reactions;;;;\n")
        # Write reactions
        maxlen = len(max(self.reactions, key=lambda x : len(x.name)))
        for rea in self.reactions:
            arrow = Reaction.directionToArrow[rea.direction]
            # Get educts and products as (coef, name, is_external) tuples
            is_external = lambda x : x[-3:].lower() == "_ex"
            educts = [(-x.coef, metabReplace[x.name], is_external(x.name))
                      for x in rea.getEducts()]
            products = [(x.coef, metabReplace[x.name], is_external(x.name))
                        for x in rea.getProducts()]

            if rea.name == self.biomass_name:
                # Save biomass for last
                biomass_data = (educts, products, arrow)
                continue

            line = "reaction;"+rea.name.ljust(maxlen)+";"
            any_ext = False
            for (_, _, is_ext) in educts+products:
                if is_ext:
                    any_ext = True
                    break
            if not any_ext:
                educts = [(coef, name, None) for (coef, name, _) in educts]
                products = [(coef, name, None) for (coef, name, _) in products]
                line += "[c]"  # cytosolic reaction
            line += (" + ".join([self._formatReactantAnnetCsv(*e+(epsilon,))
                                for e in educts]) + " %s " % arrow +
                     " + ".join([self._formatReactantAnnetCsv(*p+(epsilon,))
                                for p in products]))
            f.write(line+'\n')

        educts, products, arrow = biomass_data
        line = ("\n;Biomass Reaction;\nreaction;%s;" % self.biomass_name +
                " + ".join([self._formatReactantAnnetCsv(e[0], e[1], None,
                            epsilon) for e in educts]) + " %s " % arrow +
                " + ".join([self._formatReactantAnnetCsv(p[0], p[1], None,
                            epsilon) for p in products]))
        f.write(line+"\n\n\nThermo names;;\n\n")

    # ------ Algorithms --------------------------------------------------------

    def getBoundaryFluxes(self):
        """ identify boundary fluxes

        Returns:
        boundaryFluxes -- list of reactions which either have no products or
                          no educts
        """

        return [reaction.name for reaction in self
                if (not reaction.getEducts()) or (not reaction.getProducts())]

    def getInternalFluxes(self):
        """ identify internal reactions

        Returns list of reactions which have at least one educt and at least one
        product
        """

        return [reaction.name for reaction in self
                if reaction.getEducts() and reaction.getProducts()]


    def canBeZero(self, reactions, strict=False):
        """ for each given reaction, return whether the flux can be zero

        Keyword arguments:

        reactions -- list of reaction names
        strict    -- switch determining the behavior for reactions not present
                     in the model (True: raise KeyError, False: mark positive)

        Returns list of Boolean values indexed like reactions list
        """
        result = []
        for reaction in reactions:
            try:
                rea = self.reactions[self.reactionDict[reaction]]
                result.append(rea.lb <= 0. and rea.ub >= 0.)
            except KeyError:
                if strict:
                    raise
                else:
                    result.append(True)
        return result


    def findDuplicateReactions(self, verbose=False, epsilon=1E-8):
        """ identify duplicate reactions

        Two or more reactions in a model may be identical although they have
        different names or the reactants have a different order. Regarding the
        FBA, these reactions are equivalent and can be removed to improve the
        performance.

        Keyword arguments:

        verbose   -- if True, print debug information to the console
        epsilon   -- threshold below which a float is to be considered zero;
                     must be positive!

        Returns:
        duplicateReactions

        duplicateReactions -- set of reaction names which can be removed without
                              altering the capabilities of the model

        example/doctest:

        >>> m = MetabolicModel()
        >>> m.addReactionsFromString('''
        ... Test1 : A <=> B
        ... Test2 : A --> B
        ... Test3 : A <-- B
        ... Test4 : A <=> B
        ... Test5 : B <=> A
        ... Test6 : B <-- A''')
        >>> duplicateReactions = m.findDuplicateReactions()
        >>> m.removeReactionsByNameList(duplicateReactions)
        >>> print m
        Test1 : 1 A <=> 1 B
        Test2 : 1 A --> 1 B
        Test3 : 1 A <-- 1 B

        """

        matrix = self.getStoichiometricMatrix()
        reactionNames = array(self.getReactionNames())

        if len(matrix) <= 0:
            return []

        # add a row with reaction directions
        matrix.append([reaction.direction for reaction in self])

        # list of lists to numpy array
        matrix = array(matrix)
        _, nReactions = matrix.shape

        # make sure two identical reactions have the reactants on the same side
        for i in range(nReactions):
            nonZeros = nonzero( matrix[:,i])[0] # indices of the reactants
            # swap the reactants and the direction if necessary
            if nonZeros.any() and matrix[nonZeros[0],i] > 0:
                matrix[:,i] = matrix[:,i]*-1

        if verbose:
            print("stoichiometric matrix and reaction names:")
            print(array(matrix), "\n", reactionNames)

        # transpose matrix and create a list of lists
        matrix = [[matrix[j][i] for j in range(len(matrix))]
                                for i in range(len(matrix[0]))]

        # sort matrix and reaction names
        reactionNames = sorted(reactionNames,
                                key=lambda x: matrix[self.reactionDict[x]])
        matrix = sorted(matrix)

        if verbose:
            print("transposed and sorted stoichiometric matrix, sorted "\
                    "reaction names:")
            print(array(matrix), "\n", reactionNames)

        matrix = array(matrix)

        # search for duplicates in the sorted matrix
        duplicateReactions = [reactionNames[i] for i in range(1, nReactions)
                                if sum(abs(matrix[i-1,:]-matrix[i,:]))<=epsilon]

        if verbose:
            print("duplicate reactions:")
            print(duplicateReactions)

        return duplicateReactions


    def findDeadEnds(self, recursive=False, unbound=set(), verbose=False,
                     epsilon=1E-8):
        """ identify dead-end metabolites

        A dead end is a metabolite that cannot be produced and consumed in
        separate reactions. Reactions that contain dead-end metabolites always
        get assigned a flux of zero in flux-balance analysis.

        If recursive is True, this function recursively searches for dead ends
        and nonfunctional reactions (which in turn can induce further dead
        ends). Else, it only returns the list of dead-end metabolites.

        Keyword arguments:

        recursive -- if True, work recursively (see explanation above)
        unbound   -- set of reactions to be treated as unbounded (useful for
                     transporters, of which usually only a few are allowed at a
                     time)
        verbose   -- if True, print debug information to the console
        epsilon   -- threshold below which a float is to be considered zero;
                     must be positive!

        Returns:
        deadEnds, deadReactions

        deadEnds      -- list of dead-end metabolites (names)
        deadReactions -- list of reactions containing dead ends
                         (None if recursive is False)
        """
        nMetabolites = len(self.metabolites)
        nReactions = len(self.reactions)

        # First compile lists of sets of producing and consuming reactions for
        # every metabolite and sets of metabolites by reaction in which they
        # participate
        # producingReactions and consumingReactions take bounds into account
        producingReactions = [set() for _ in range(nMetabolites)]
        consumingReactions = [set() for _ in range(nMetabolites)]
        metabolitesByReaction = [set() for _ in range(nReactions)]

        allDeadReactions = set()

        for j in range(nReactions):
            rea = self.reactions[j]
            isUnbounded = rea.name in unbound
            canBePos = rea.ub > epsilon
            canBeNeg = rea.lb < -epsilon
            if not isUnbounded and not canBePos and not canBeNeg:
                # Mark reaction as dead if its flux is constrained to zero
                allDeadReactions.add(j)
                continue

            for coef, name in rea:
                i = self.metaboliteDict[name]

                if abs(coef) > epsilon:
                    metabolitesByReaction[j].add(i)
                    if coef > 0.:
                        # Coefficient is positive, i.e. if the flux is
                        # positive, the metabolite is produced
                        if canBePos or isUnbounded:
                            producingReactions[i].add(j)
                        if canBeNeg or isUnbounded:
                            consumingReactions[i].add(j)
                    else:
                        # Coefficient is negative, i.e. if the flux is
                        # positive, the metabolite is consumed
                        if canBePos or isUnbounded:
                            consumingReactions[i].add(j)
                        if canBeNeg or isUnbounded:
                            producingReactions[i].add(j)

        if verbose:
            nDeadReactions = len(allDeadReactions)
            if nDeadReactions:
                print ("%u reactions have a flux that is restricted to zero:" %
                       nDeadReactions)
                print("\n".join(sorted(self.reactions[i].name for i in
                                       allDeadReactions)))
            else:
                print("There are no reactions with a flux restricted to zero.")
            print()
            print("Producing and consuming reactions for all metabolites:")
            for i in range(nMetabolites):
                print ("%s (%u, %u)" % (self.metabolites[i],
                       len(producingReactions[i]), len(consumingReactions[i])))
            print()

        # Total reactions is the combined set of consuming and producing
        # reactions
        reactionsByMetabolite = [consumingReactions[i] |
            producingReactions[i] for i in range(nMetabolites)]

        isDeadEnd = [False]*nMetabolites

        deadReactions = 1   # dummy value (to enter loop)
        while deadReactions:
            deadReactions = set()

            # 1. Identify dead ends, check for dead reactions

            for i in range(nMetabolites):
                if isDeadEnd[i]:
                    continue  # Don't test known dead ends again

                if (not producingReactions[i] or not consumingReactions[i] or
                    len(reactionsByMetabolite[i]) <= 1):

                    # Metabolite i is a dead end
                    isDeadEnd[i] = True

                    # All reactions that involve this metabolite are also "dead"
                    if verbose:
                        newlyDeadReactions = set()
                        for j in reactionsByMetabolite[i]:
                            deadReactions.add(j)
                            newlyDeadReactions.add(self.reactions[j].name)

                        print("%s prod: %s consu: %s is a dead end (reactions: "
                              "%s)" % (self.metabolites[i],
                                       [self.reactions[j].name
                                        for j in producingReactions[i]],
                                       [self.reactions[j].name for j in
                                        consumingReactions[i]],
                                       ", ".join(newlyDeadReactions)))
                    else:
                        for j in reactionsByMetabolite[i]:
                            deadReactions.add(j)

            if not recursive:
                deadReactions = None
                break

            # 2. Update producingReactions & consumingReactions accordingly

            for j in deadReactions:
                if verbose:
                    print(("Fixing reaction %s (metabolites:" %
                           self.reactions[j].name), end=' ')

                # Remove dead reaction from sets of producing, consuming, and
                # total reactions of every involved metabolite
                for i in metabolitesByReaction[j]:
                    if not isDeadEnd[i]:
                        tmpSet = set((j,))
                        producingReactions[i] -= tmpSet
                        consumingReactions[i] -= tmpSet
                        reactionsByMetabolite[i] -= tmpSet

                allDeadReactions.add(j)

                if verbose:
                    print("%s)" % ", ".join([self.metabolites[i] for i in
                                metabolitesByReaction[j] if not isDeadEnd[i]]))

        # Replace indices with names
        deadEnds = [self.metabolites[i] for i in range(nMetabolites)
                    if isDeadEnd[i]]
        deadReactions = [self.reactions[j].name for j in allDeadReactions]

        if verbose:
            print()
            if recursive:
                print ("Producing and consuming reactions after removal of all "
                       "dead ends:")
                for i in range(nMetabolites):
                    print ("%s (%u, %u)" % (self.metabolites[i],
                        len(producingReactions[i]), len(consumingReactions[i])))
                print()

        return deadEnds, deadReactions


    def getActiveReactions(self, solution, threshold=1E-6, messages=[],
                           coefCutoff=1E-12):
        """ identify the reactions that have an absolute flux above the given
            threshold and all metabolites involved in these reactions

        Keyword arguments:

        solution   -- vector of flux values given as either
                        a) list of values with same order as reactions list
                        b) dictionary {reaction name : value}
        threshold  -- flux threshold above which a reaction is considered
                      "active"
        messages   -- list in which to store warnings and info messages (as
                      (level, msg) pairs)
        coefCutoff -- cutoff below which a stoichiometric coefficient is
                      considered zero

        If solution is given as a dict, all reactions which don't have an entry
        in solution, are considered "inactive". If solution is given as another
        iterable, its length must be the same as the number of reactions in the
        model.

        Returns:
        activeReactions, activeMetabolites

        activeReactions   -- list of active reactions
        activeMetabolites -- dict {name : #occurrences} listing the number of
                             active reactions in which each "active" metabolite
                             occurs
        """
        if isinstance(solution, dict):
            flux = solution
        else:
            if len(solution) != len(self.reactions):
                raise ValueError("Length of solution vector and number of "
                                 "reactions in the model disagree")
            flux = dict(list(zip(self.getReactionNames(), solution)))
        activeMetabolites = {}
        activeReactions = []
        for reaname in flux:
            if reaname not in self.reactionDict:
                msg = ("Warning: Reaction '%s' from solution not found in "
                       "model." % reaname)
                messages.append((Verbosity.WARNING, msg))
                continue

            if abs(flux[reaname]) > threshold:
                activeReactions.append(reaname)
                rea_index = self.reactionDict[reaname]

                for coef, name in self.reactions[rea_index]:
                    if abs(coef) > coefCutoff:
                        if name in activeMetabolites:
                            activeMetabolites[name] += 1
                        else:
                            activeMetabolites[name] = 1

        return activeReactions, activeMetabolites



from setuptools import setup
from setuptools.command import sdist

sdist.walk_revctrl = lambda **kwargs: []


with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="metano",
    version="1.3.0",
    description="metano - A Toolkit for Metabolic Network Analysis and "
    "Optimization",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Julia Helmecke and Alexander Riemer",
    author_email="julia.helmecke@tu-bs.de",
    url='http://metano.tu-bs.de/',
    packages=["metano"],
    package_dir={"metano": "src"},
    package_data={
        "metano": ["example/*.txt", "example/*.sh", "example/*.ini"]
    },
    entry_points={
        "console_scripts": [
            'deadends.py = metano.deadends:main',
            'fba.py = metano.fba:main',
            'fva.py = metano.fva:main',
            'moma.py = metano.moma:main',
            'knockout.py = metano.knockout:main',
            'sbmlparser.py = metano.sbmlparser:main',
            'sbmlwriter.py = metano.sbmlwriter:main',
            'splitratios.py = metano.splitratios:main',
            'mfba.py = metano.mfba:main',
            'mfm.py = metano.mfm:main',
            'modelassert.py = metano.modelassert:main',
            'moment.py = metano.moment:main',
            'to_check_constraints.py = metano.to_check_constraints:main',
            'to_diff_solutions.py = metano.to_diff_solutions:main',
            'to_evaluate_objective.py = metano.to_evaluate_objective:main',
            'to_fix_names.py = metano.to_fix_names:main',
            'to_list_active_metab.py = metano.to_list_active_metab:main',
            'to_rea2m.py = metano.to_rea2m:main',
            'to_scatterplot.py = metano.to_scatterplot:main',
        ]
    },
    classifiers=[
        "License :: OSI Approved :: GNU General Public License (GPL)",
        "Programming Language :: Python :: 3",
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    keywords="genome-scale metabolic models flux balance analysis fba",
    license="GPL",
    install_requires=[
        "numpy",
        "swiglpk",
        "pymprog>=1.0",
        "ecos",
        "scipy",
        "openopt",
        "cvxpy",
        "cvxopt",
    ]
)



"""
This module defines the class ReactionParser, a parser for reaction files.


This file is part of metano.
Copyright (C) 2010-2019 Alexander Riemer, Julia Helmecke
Braunschweig University of Technology,
Dept. of Bioinformatics and Biochemistry

metano is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

metano is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with metano.  If not, see <http://www.gnu.org/licenses/>.
"""
from __future__ import absolute_import

from builtins import map
from builtins import object
from metano.defines import Verbosity, ReaFileStruc


def _splitAtPlus(s):
    """ split the string at plus signs that are preceded or followed by
        whitespace
    """
    tokens = list(map(str.strip, s.split(' +')))
    tokensList = [x.split('\t+') for x in tokens]
    tokens = []
    for t in tokensList:
        tokens += list(map(str.strip, t))
    tokensList = [x.split('+ ') for x in tokens]
    tokens = []
    for t in tokensList:
        tokens += list(map(str.strip, t))
    tokensList = [x.split('+\t') for x in tokens]
    tokens = []
    for t in tokensList:
        tokens += list(map(str.strip, t))
    return tokens


class ReactionParser(object):
    """Parser for reaction files.
    It reads lines of the format

    <reaction> : <compound> [+ <compound>...] -->|<--|<=>
                 <compound> [+ <compound>...]

    and builds the stoichiometric matrix of the corresponding metabolic network,
    along with dictionaries of reactions and compounds, which relate names to
    the corresponding rows or columns in the matrix, and a list of reversibility
    flags for all reactions.
    """

    format_hint = ("Format is\n<reaction> %s <compound> [+ <compound>...] "
                   "%s|%s|%s <compound> [+ <compound>...]" %
                   (ReaFileStruc.delimiter, ReaFileStruc.arrowIrr,
                    ReaFileStruc.arrowFlip, ReaFileStruc.arrowRev))

    def __init__(self, skipCommentLines=False):
        """ initialize the ReactionParser
        """
        self.clear()
        self.skipCommentLines = skipCommentLines

    def clear(self):
        """ clear all internal variables, warnings and info messages
        """
        # List of warnings and info messages generated during parsing,
        # elements are pairs (level, msg)
        self.messages = []
        # If true, don't assign rows starting with a hash sign (#) to the
        # reaction below the comment
        self.skipCommentLines = False

    def getMessages(self, level=Verbosity.INFO):
        """ return a list of all messages at or above the given level of
            severity
            Levels are defined in defines.py.
        """
        return [x[1] for x in self.messages if x[0] <= level]

    def parse(self, filename):
        """ parse the reaction file identified by filename
            -- wrapper for parseByHandle()
        """
        with open(filename) as f:
            return self.parseByHandle(f)

    def parseByHandle(self, f):
        """ parse the reaction file given as a file object

        Returns:
        list of (name, educts, arrow, products, comment) tuples with

        name     -- reaction name
        educts   -- list of (coef, name) tuples for left-hand-side metabolites
        arrow    -- reaction arrow (<=>, -->, or <--)
        products -- list of (coef, name) tuples for right-hand-side metabolites
        comment  -- an optional comment for the reaction
        """
        line_no = 0
        result = []
        comment = ""
        metabSet, reactionSet = set(), set()
        for line in f:
            line_no += 1
            if line.startswith(ReaFileStruc.commentSign):
                continue

            # Everything after comment sign ('#') is comment
            comment_pos = line.find(ReaFileStruc.commentSign)
            if comment_pos > 0 or (comment_pos == 0 and
                                   not self.skipCommentLines):
                if comment:
                    # Append to previous line of comment
                    comment += '\n'
                comment += line[comment_pos+1:].strip()
            if comment_pos >= 0:
                line = line[:comment_pos].strip()

            if line == "" or line.isspace():
                if comment and comment_pos < 0:
                    comment += '\n'
                continue  # Skip blank lines
            else:
                # Current line contains a reaction -> finish comment
                reaction_comment = comment
                comment = ""

            # Split line at delimiter
            delimiter_pos = line.find(ReaFileStruc.delimiter)
            if delimiter_pos < 0:
                msg = self.error(line_no)
                raise SyntaxError(msg)

            rea_name = line[:delimiter_pos].strip()
            if rea_name == "":  # Reaction name must not be empty
                msg = self.error(line_no)
                raise SyntaxError(msg)

            if rea_name in reactionSet:
                msg = self.error(line_no, "Duplicate definition of reaction " +
                                 rea_name, False)
                raise SyntaxError(msg)
            if len(rea_name.split(None, 1)) > 1:
                msg = ("Warning: Reaction name '%s' contains space/tab "
                       "(reaction file, line %u)" % (rea_name, line_no))
                self.messages.append((Verbosity.WARNING, msg))
            reactionSet.add(rea_name)

            rest = line[delimiter_pos+len(ReaFileStruc.delimiter):]
            if rest.find(ReaFileStruc.delimiter) >= 0:
                msg = self.error(line_no, "Extra delimiter ('%s')" %
                                 ReaFileStruc.delimiter)
                raise SyntaxError(msg)

            # Split rest at arrow
            for arrow in ReaFileStruc.arrow_prop:
                arrow_pos = rest.find(arrow)
                if arrow_pos >= 0:
                    arrowprop = ReaFileStruc.arrow_prop[arrow]
                    break

            if arrow_pos < 0:
                msg = self.error(line_no)
                raise SyntaxError(msg)

            # Get reactants and products
            educts_str = rest[:arrow_pos]
            products_str = rest[arrow_pos+arrowprop.length:]

            line_minus_arrow = educts_str + '\n' + products_str
            # ('\n' is never part of arrow string)
            for arrow_ in ReaFileStruc.arrow_prop:
                if line_minus_arrow.find(arrow_) >= 0:
                    msg = self.error(line_no, "Extra arrow")
                    raise SyntaxError(msg)

            educts = _splitAtPlus(educts_str)
            products = _splitAtPlus(products_str)

            # Get reactants as (coefficient, name) pairs
            educts = self.process_metabolites(educts, line_no)
            products = self.process_metabolites(products, line_no)

            # Check for duplicate metabolites in reaction
            metabNames = set()
            for name in [x[1] for x in educts+products]:
                if name in metabNames:
                    msg = ("Warning: Duplicate metabolite '%s' in line %u in "
                           "reaction file." % (name, line_no))
                    self.messages.append((Verbosity.WARNING, msg))
                else:
                    metabNames.add(name)
                    metabSet.add(name)

            result.append((rea_name, educts, arrow,
                           products, reaction_comment))

        msg = ("Info: The metabolic network has %u reactions and %u metabolites"
               ".") % (len(reactionSet), len(metabSet))
        self.messages.append((Verbosity.INFO, msg))
        return result

    def process_metabolites(self, metab_strings, line_no):
        """ process a list of metabolites in a reaction (reactants or products)

        Keyword arguments:

        metab_strings -- list of metabolites given as strings in the form
                             "<coefficient> <metabolite>"
        line_no       -- line number in file

        Returns: list of (coefficient, name) pairs
        """
        result = []
        for metab_string in metab_strings:
            if metab_string == "":  # Skip empty strings
                continue

            entry = metab_string.split(None, 2)
            if len(entry) > 2:
                msg = self.error(line_no, "Metabolite names must not contain "
                                 "spaces/tabs.\n Illegal name encountered")
                raise SyntaxError(msg)
            elif len(entry) == 1:
                name = entry[0]
                coefficient = 1.
            else:
                # Split entry into coefficient and name
                try:
                    coefficient = float(entry[0])
                except ValueError:
                    msg = self.error(line_no, "Metabolite names must not "
                                     "contain spaces/tabs.\n Illegal name "
                                     "encountered", False)
                    raise SyntaxError(msg)
                name = entry[1]

            if coefficient <= 0.:
                msg = self.error(line_no, "Negative or zero coefficient "
                                 "encountered", False)
                raise SyntaxError(msg)

            result.append((coefficient, name))
        return result

    def error(self, line_no=0, msg="Malformed reaction", hint=True):
        """ generate an error message

        Keyword arguments:

        line_no     -- the line number, 0 if not applicable
        msg         -- error message
        hint        -- show hint on expected format of reactions (True/False)
        """
        s = msg
        if line_no > 0:
            s += " in line %u" % line_no
        if hint:
            s += "\n"
            s += self.format_hint
        return s

    def fixfile(self, filename, outfilename):
        """ fix reaction file by replacing spaces in reaction and metabolite
            names with underscores and converting metabolite names to lowercase

        Keyword arguments:

        filename    -- input file
        outfilename -- name of output file to be written
        """
        with open(filename) as f:
            with open(outfilename, 'w') as outf:
                self.fixfileByHandles(f, outf)

    def fixfileByHandles(self, f, outf):
        """ fix reaction file by replacing spaces in reaction and metabolite
            names with underscores and converting metabolite names to lowercase

        Keyword arguments:

        f    -- file handle of input file (must be open for reading)
        outf -- file handle of output file (must be open for writing)
        """
        line_no = 0
        reaNamesFixed = 0
        metabDict = {}  # dictionary {name : new name}
        for line in f:
            line_no += 1

            # Everything after comment sign ('#') is comment
            comment_pos = line.find(ReaFileStruc.commentSign)
            if comment_pos >= 0:
                # Get comment with comment sign
                comment = line[comment_pos:].rstrip()
            else:
                comment = ""
            line = line[:comment_pos]
            if line == "" or line.isspace():
                outf.write(comment+'\n')
                continue  # Skip blank lines

            # Split line at delimiter
            delimiter_pos = line.find(ReaFileStruc.delimiter)
            if delimiter_pos < 0:
                msg = self.error(line_no)
                raise SyntaxError(msg)

            # Fix reaction name
            rea_name = line[:delimiter_pos].strip()
            if rea_name == "":  # Reaction name must not be empty
                msg = self.error(line_no)
                raise SyntaxError(msg)

            rea_name_split = rea_name.split()
            if len(rea_name_split) > 1:
                reaNamesFixed += 1
                rea_name_new = '_'.join(rea_name_split)
                prefix = (rea_name_new + " %s " %
                          ReaFileStruc.delimiter)
                msg = "Reaction '%s' --> '%s'" % (rea_name, rea_name_new)
                self.messages.append((Verbosity.DEBUG, msg))
            else:
                prefix = line[:delimiter_pos+len(ReaFileStruc.delimiter)] + ' '

            rest = line[delimiter_pos+len(ReaFileStruc.delimiter):]
            if rest.find(ReaFileStruc.delimiter) >= 0:
                msg = self.error(line_no, "Extra delimiter ('%s')" %
                                 ReaFileStruc.delimiter)
                raise SyntaxError(msg)

            # Split rest at arrow
            for arrow in ReaFileStruc.arrow_prop:
                arrow_pos = rest.find(arrow)
                if arrow_pos >= 0:
                    arrowprop = ReaFileStruc.arrow_prop[arrow]
                    arrowStr = ' ' + arrow + ' '
                    break

            if arrow_pos < 0:
                msg = self.error(line_no)
                raise SyntaxError(msg)

            # Get reactants and products
            educts_str = rest[:arrow_pos]
            products_str = rest[arrow_pos+arrowprop.length:]

            line_minus_arrow = educts_str + '\n' + products_str
            # ('\n' is never part of arrow string)
            for arrow_ in ReaFileStruc.arrow_prop:
                if line_minus_arrow.find(arrow_) >= 0:
                    msg = self.error(line_no, "Extra arrow")
                    raise SyntaxError(msg)

            educts = _splitAtPlus(educts_str)
            products = _splitAtPlus(products_str)

            # Fix metabolite names (replace spaces with underscores)
            educts = self.fix_metab_names(educts, metabDict)
            products = self.fix_metab_names(products, metabDict)

            # Write line to output file
            outf.write(prefix + " + ".join(educts) + arrowStr +
                       " + ".join(products) + '\n')
        msg = ("Fixed %u reaction name(s) and %u metabolite name(s)." %
               (reaNamesFixed, len([key for key in metabDict
                                    if key != metabDict[key]])))
        self.messages.append((Verbosity.INFO, msg))

    def fix_metab_names(self, metab_strings, metabDict):
        """ process a list of metabolites; called internally by fixfileByHandles

        This function replaces whitespace in metabolite names with underscores
        and converts metabolite names to all lowercase. It tries to interpret
        the first "word" of a given metabolite name containing spaces as the
        stoichiometric coefficient.

        Keyword arguments:

        metab_strings -- list of metabolites given as strings in the form
                         "<coefficient> <metabolite>" (where metabolite may
                         still contain whitespace characters)
        metabDict     -- dictionary {name : fixed name}

        Returns:

        names_fixed   -- list of fixed "<coefficient> <metabolite>" strings,
                         where metabolite does not contain any whitespace
                         characters
        """
        newNames = set(metabDict.values())
        names_fixed = []
        for metab_string in metab_strings:
            if not metab_string:
                continue

            entry = metab_string.lower().split(None, 1)
            if len(entry) > 1:
                # Try to interpret first part as coefficient
                try:
                    float(entry[0])
                    names_fixed.append(entry[0] + ' ' +
                                       self.getFixedMetabName(entry[1], metabDict, newNames))
                except ValueError:
                    names_fixed.append(self.getFixedMetabName(
                        metab_string.lower(), metabDict, newNames))
            else:
                names_fixed.append(entry[0])
        return names_fixed

    def getFixedMetabName(self, name, metabDict, newNames):
        """ return fixed metabolite name; called internally by fix_metab_names()

        Keyword arguments:

        name        -- metabolite name to fix
        metabDict   -- dictionary {name : fixed name}
        newNames    -- set of metabDict.values()
        """
        if name in metabDict:
            return metabDict[name]
        fixed_name = name.replace(' ', '_')
        while fixed_name in newNames:
            fixed_name += '_'
        metabDict[name] = fixed_name
        newNames.add(fixed_name)
        if name != fixed_name:
            msg = "Metabolite '%s' --> '%s'" % (name, fixed_name)
            self.messages.append((Verbosity.DEBUG, msg))
        return fixed_name
